from reportlab.platypus import Paragraph, Table, Spacer, PageBreak, XPreformatted
from reportlab.lib.units import inch
from reportlab.lib.styles import getSampleStyleSheet
from pygments import highlight
from pygments.lexers import PythonLexer
from pygments.formatters import get_formatter_by_name
from pygments.styles import get_style_by_name
from pygments.token import STANDARD_TYPES, Token
import numpy as np
from pathlib import Path
import re

from .settings import *
from ..settings import DocsType
from ...settings import Sign, ROOT_SOURCE
from ...pygments import SyntaxLexer, StyleLexer

class SourcesSection:
    
    data: list
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
        
    def __init__(self, data: list):
        self.data = data
        
    def highlight_python_code(self, item):
        # format text using pygments
        lexer = SyntaxLexer()
        style = StyleLexer

        # format text with a lexer
        formatter = get_formatter_by_name('html', style=style) 
        code = highlight(item.code, lexer, formatter)
        
        # add code lines
        lines = code.split(Sign.NEWLINE)
        for l in range(len(lines)-2):  # the last two lines are always empty
            lineno = str(l+1)
            lines[l] = Target(f"{item.target}_{lineno}", f" {lineno:5s}{lines[l]}")
        code = Sign.NEWLINE.join(lines)

        # replace standard CSS classes with explicit text formatting
        token_settings = dict(style)
        for token, class_name in STANDARD_TYPES.items():
            token_style = token_settings[token]
            attributes = []
            if token_style['color']:
                attributes.append(f"color=\"#{token_style['color']}\"")
            attributes = " ".join(attributes)
            code = code.replace(f"class=\"{class_name}\"", attributes)   
        
        # replace DIP specific CSS classes with explicit text formatting
        def replace(m):
            token = getattr(Token.DIP,m.group(1))
            if token not in token_settings:
                return f">{m.group(2)}"
            else:
                token_style = token_settings[token]
                attributes = []
                if token_style['color']:
                    attributes.append(f"color=\"#{token_style['color']}\"")
                attributes = " ".join(attributes)
                if token_style['bold']:
                    return f"{attributes}><b>{m.group(2)}</b>"
                else:
                    return f"{attributes}>{m.group(2)}"
        code = re.sub("class=\" \-DIP \-DIP\-([a-zA-Z]*)\">(.*?)(?=</span)", replace, code)

        return code
    
    def parse_item(self, item):
        
        current_path = Path(item.path)
        
        CODE_STYLE = getSampleStyleSheet()["Code"]
        TABLE_STYLE = [                       
            ('GRID',       (0,1), (-1,-1),  0.5, PALETTE['prop_name']),
            ('BACKGROUND', (0,0), (1,0),   PALETTE['node_name']),  
            ('BACKGROUND', (0,1), (0,-1),   PALETTE['prop_name']),  
            ('BACKGROUND', (1,1), (1,-1),    PALETTE['prop_value']),  
            #('SPAN',       (0,-1), (1,-1)     ),   
        ]


        blocks = []
        p = Paragraph(Target(item.target) + item.name)

        data = [   
            [p,  ''],
            ['File:', current_path.name],
        ]
        if ROOT_SOURCE not in item.name:
            src = Paragraph(Link(item.link_source, f"{item.parent[0]}:{item.parent[1]}"))
            data.append(['Source:', src ])
        t = Table(data, style=TABLE_STYLE, hAlign='LEFT', colWidths=ColumnWidths([0.2, 0.8]))
        t.keepWithNext = True
        blocks.append(t)

        if not ROOT_SOURCE in item.name:
            code = self.highlight_python_code(item)
            c = XPreformatted(code, CODE_STYLE) 
            blocks.append(Spacer(1,0.1*inch))
            blocks.append(c)
        
        return blocks
        
    def parse(self):
        blocks = []
        blocks.append(Paragraph(Title(f"List of sources"), H2))
        for item in self.data:
            blocks.append(Spacer(1,0.1*inch))
            blocks += self.parse_item(item)
            blocks.append(Spacer(1,0.1*inch))
        blocks.append(Spacer(1,0.2*inch))
        return blocks


from dataclasses import dataclass, field
from typing import List, Dict, Union
import numpy as np

from ..settings import Sign

@dataclass
class Parent:
    indent: int 
    name: str   
    
@dataclass
class HierarchyList:
    parents: List[Parent] = field(default_factory = list)
    
    def register(self, node, excluded):
        """ Register new node to the hierarchy

        :param node: Node that should be added
        :param list excluded: List of node keywords that should be excluded from the parents
        """
        if node.name is not None and node.keyword not in excluded:
            while self.parents and node.indent<=self.parents[-1].indent:
                self.parents.pop()
            self.parents.append(Parent(node.indent, node.name))
            node.name = Sign.SEPARATOR.join([parent.name for parent in self.parents])


import numpy as np

from .settings import *
from .quantity import Quantity

class Constant:

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass
    
    def __new__(cls, unit=None):
        if unit:
            return Quantity(1,f"[{str(unit)}]")
        else:
            return object.__new__(cls)
    
    def __getattr__(self, unit):
        return Quantity(1,f"[{str(unit)}]")

    def __str__(self):
        return Constant._list()
        
    def __rep__(self):
        return Constant._list()

    @staticmethod
    def _list():
        sw, uw, dw = 8, 20, 20
        text = "Constants\n\n"
        text += f"{'Symbol':{sw}s} | {'Unit':{uw}s} | {'Definition':{dw}s}\n"
        text += "-"*sw+"-+-"+"-"*uw+"-+-"+"-"*dw+"\n"
        for symbol, unit in UNIT_STANDARD.items():
            if symbol.startswith('['):
                text += f"{symbol:{sw}s} | {unit['name']:{uw}s} | {unit['definition']:{dw}s}\n"
        return text
        
    @staticmethod
    def list():
        print(Constant._list())
            


import sys, os
import pytest
import numpy as np
sys.path.insert(0, 'src')
from pathlib import Path
from inspect import getframeinfo, stack
from dataclasses import asdict

from scinumtools.dip import DIP
from scinumtools.dip.settings import ROOT_SOURCE, FILE_SOURCE, STRING_SOURCE

def test_root_source():
    
    with DIP() as dip:

        # test if root source is correct
        assert ROOT_SOURCE in dip.source[0]
        assert dip.source[0] in dip.env.sources
        rs = dip.env.sources[dip.source[0]]
        assert rs.path == __file__
        
def test_file_sources():
        
    with DIP() as dip:
        
        source_fa = f"{dip.name}_{FILE_SOURCE}1"
        source_fb = f"{dip.name}_{FILE_SOURCE}2"
        file_sources = {
            source_fa: dict(
                name   = source_fa,
                path   = str(Path(__file__).parent/"examples/source_fa.dip"),   # absoute path
                nodes  = None,
                sources = None,
            ),
            source_fb: dict(
                name   = source_fb,
                path   = "examples/source_fb.dip",   # relative path
                nodes  = None,
                sources = None,
            )
        }
        
        # import code from file
        dip.add_file(file_sources[source_fa]['path'])
        dip.add_file(file_sources[source_fb]['path'])
        
        # get the line numbers of the callings above
        caller = getframeinfo(stack()[0][0])
        file_sources[source_fa]['parent'] = (dip.source[0], caller.lineno-4)
        file_sources[source_fb]['parent'] = (dip.source[0], caller.lineno-3)
        
        # get code from files
        with open(file_sources[source_fa]['path'],'r') as f:  # absolute path
            file_sources[source_fa]['code'] = f.read()
        file_sources[source_fb]['path'] = str(Path(__file__).parent/file_sources[source_fb]['path'])
        with open(file_sources[source_fb]['path'],'r') as f:  # relative path
            file_sources[source_fb]['code'] = f.read()
    
        # test if file sources have correct data
        for source_name, match in file_sources.items():
            assert source_name in dip.env.sources
            assert asdict(dip.env.sources[source_name]) == match
    
        # test if nodes have correct sources and line numbers
        nodes = [
            ('runtime.timestep', source_fa,  3),
            ('size',             source_fb,  1),
        ]
        env = dip.parse()
        for name, source, lineno in nodes:
            node = env.nodes.query(name)[0]
            assert node.source == (source, lineno)

def test_string_sources():

    with DIP() as dip:
        
        source_sa = f"{dip.name}_{STRING_SOURCE}1"
        source_sb = f"{dip.name}_{STRING_SOURCE}2"
        string_sources = {
            source_sa: dict(
                name   = source_sa,
                path   = __file__,
                code   = """name str = 'John Smith'
                age int = 23""",
                nodes  = None,
                sources = None,
            ),
            source_sb: dict(
                name   = source_sb,
                path   = __file__,
                code   = """salary float = 60000""",
                nodes  = None,
                sources = None,
            ),
        }
        
        # import code from string
        dip.add_string(string_sources[source_sa]['code'])
        dip.add_string(string_sources[source_sb]['code'])
  
        # get the line numbers of the callings above
        caller = getframeinfo(stack()[0][0])
        string_sources[source_sa]['parent'] = (dip.source[0], caller.lineno-4)
        string_sources[source_sb]['parent'] = (dip.source[0], caller.lineno-3)

        
        # test if string sources have correct data
        for source_name, match in string_sources.items():
            assert source_name in dip.env.sources
            assert asdict(dip.env.sources[source_name]) == match
       
        # test if nodes have correct sources and line numbers
        nodes = [
            ('name',             source_sa,  1),
            ('salary',           source_sb,  1),
        ]
        env = dip.parse()
        for name, source, lineno in nodes:
            node = env.nodes.query(name)[0]
            assert node.source == (source, lineno)

def test_explicit_sources():
    
    with DIP() as dip:
        
        source_file = "examples/source_fa.dip"
        
        dip.add_source("explicit", source_file)   # explicitely added
        env = dip.parse()
        
        # get line number
        caller = getframeinfo(stack()[0][0])
        
        # read code from the source file
        source_path = str(Path(__file__).parent/source_file)
        with open(source_path,'r') as f:  # relative path
            source_code = f.read()

        # check the sources
        assert 'explicit' in env.sources
        source = env.sources['explicit']
        assert source.name == 'explicit'
        assert source.path == source_path
        assert source.code == source_code
        assert ROOT_SOURCE in source.parent[0]
        assert source.parent[1] == caller.lineno-4
        
        # test if nodes have correct sources and line numbers
        nodes = [
            ('explicit?runtime.timestep', 'explicit',  3),
        ]
        for name, source, lineno in nodes:
            node = env.request(name)[0]
            assert node.source == (source, lineno)

def test_inline_sources():
    
    with DIP() as dip:
        
        source_file = "examples/source_fa.dip"
        
        dip.add_string(f"""
        $source inline = {source_file}            # source added in code
        name str = 'John Smith'
        """)
        env = dip.parse()
        
        # read code from the source file
        source_path = str(Path(__file__).parent/source_file)
        with open(source_path,'r') as f:  # relative path
            source_code = f.read()

        # check the sources
        source_name = f"{dip.name}_{STRING_SOURCE}1"
        assert source_name in env.sources
        assert 'inline' in env.sources
        source = env.sources['inline']
        assert source.name == 'inline'
        assert source.path == source_path
        assert source.code == source_code
        assert source.parent == (source_name, 1)

        # test if nodes have correct sources and line numbers
        nodes = [
            ('inline?runtime.timestep', 'inline',  3),
        ]
        for name, source, lineno in nodes:
            node = env.request(name)[0]
            assert node.source == (source, lineno)


from enum import Enum
MAGNITUDE_PRECISION = 1e-7

# Fraction input type
class Norm(Enum):
    NUMBER          = 0    # number
    NUMBER_FRACTION = 1    # number fraction
    MASS_FRACTION   = 2    # mass fraction

# Standard units used by this submodule
class Units:
    ATOMIC_MASS    = "Da"
    MATERIAL_MASS  = "g"
    NUMBER_DENSITY = "cm-3"
    MASS_DENSITY   = "g/cm3"
    FRACTION       = "%"

from .element import Element
from .substance import Substance
from .substance_solver import SubstanceSolver
from .material import Material
from .material_solver import MaterialSolver



import numpy as np

from ...solver import *
from ...units import UnitEnvironment
from ..environment import Environment
from ..nodes.parser import Parser
from ..datatypes import NumberType, BooleanType
from ..settings import Keyword, Sign

class LogicalSolver:

    env: Environment
    
    def __init__(self, env:Environment=None, **kwargs):
        if env:
            self.env = env
        else:
            self.env = Environment()
            
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
    
    def _eval_node(self, expr):
        expr = expr.strip()
        if expr=='':
            return None
        flags = []
        if expr[0]==Sign.DEFINED:
            flags.append('defined')
            expr = expr[1:]
        # parse node from the code
        kwargs = {'code': expr, 'source':None}
        p = Parser(keyword='node',**kwargs)
        p.part_value()
        if p.value_ref:   # import existing node
            nodes = self.env.request(p.value_ref, count=[0,1])
            if len(nodes)==1:
                if 'defined' in flags:
                    node = BooleanType(True)
                else:
                    node = nodes[0].value
            elif len(nodes)==0:
                if 'defined' in flags:
                    node = BooleanType(False)
                else:
                    raise Exception('Missing reference:', expr)
        elif p.value_raw==Keyword.TRUE:
            node = BooleanType(True)
        elif p.value_raw==Keyword.FALSE:
            node = BooleanType(False)
        else:            # create anonymous node
            p.part_units()
            node = NumberType(p.value_raw, p.units_raw)
        return node

    def solve(self, expr):
        with UnitEnvironment(self.env.units):
            operators = {
                'par': OperatorPar,        # should be the last of parenthesis operators
                'eq': OperatorEq, 'ne': OperatorNe,
                'not': CustomNot,          # needs to be after OperatorNe
                'le': OperatorLe, 'ge': OperatorGe,
                'lt': OperatorLt, 'gt': OperatorGt,
                'and': CustomAnd, 'or': CustomOr, 
            }
            with ExpressionSolver(self._eval_node, operators) as es:
                return es.solve(expr)
                

class CustomNot(OperatorNot):
    symbol: str = Sign.NEGATE

class CustomAnd(OperatorAnd):
    
    def operate_binary(self, tokens):
        left, right = tokens.get_left(), tokens.get_right()
        if isinstance(left, (bool, np.bool_)):
            left = BooleanType(left)
        if isinstance(right, (bool, np.bool_)):
            right = BooleanType(right)
        tokens.put_left(left.logical_and(right))
        
class CustomOr(OperatorOr):
    
    def operate_binary(self, tokens):
        left, right = tokens.get_left(), tokens.get_right()
        if isinstance(left, (bool, np.bool_)):
            left = BooleanType(left)
        if isinstance(right, (bool, np.bool_)):
            right = BooleanType(right)
        tokens.put_left(left.logical_or(right))


from typing import Union, List
import numpy as np

from .type import Type

class BooleanType(Type):
    value: Union[bool,list]
    dtype = bool

    def __str__(self):
        return f"BooleanType({self.value})"
    
    def __repr__(self):
        return f"BooleanType({self.value})"
        
    def __init__(self, value, **kwargs):
        if isinstance(value, BooleanType):
            kwargs['value'] = value.value
        else:
            kwargs['value'] = value
        if isinstance(kwargs['value'], np.ndarray):
            kwargs['value'] = kwargs['value'].tolist()
        super().__init__(**kwargs)

    def __eq__(self, other):
        if isinstance(other, bool):
            return self.value == other
        else:
            return self.value == other.value
        
    def logical_and(self, other):
        return BooleanType(self.value and other.value)
        
    def logical_or(self, other):
        return BooleanType(self.value or other.value)
        
    def logical_not(self):
        return BooleanType(not self.value)
        


from .settings import *
from .quantity import Quantity 

class SystemOfUnits:
    def __init__(self, prefix):
        self.prefix = prefix
        self.names = list(QUANTITY_LIST.name)
        self.symbols = list(QUANTITY_LIST.symbol)
    def __getattr__(self, quantity):
        idx = self.names.index(quantity)
        symbol = self.symbols[idx]
        def system(x=1):
            unit = f"{SYMBOL_SYSTEM_UNIT}{self.prefix}{symbol}"
            return Quantity(x, unit)
        return system
    def __getitem__(self, quantity):
        idx = self.names.index(quantity)
        symbol = self.symbols[idx]
        unit = f"{SYMBOL_SYSTEM_UNIT}{self.prefix}{symbol}"
        return unit

SI    = SystemOfUnits('S')
AU    = SystemOfUnits('A')
CGS   = SystemOfUnits('C')



"""
This script works in a similar way as pytests, but it is used to build code parts or generate tables.

Example of use:

python3 tools/generator/generator.py
python3 tools/generator/generator.py build_docs
python3 tools/generator/generator.py build_docs::build_temperature_units
"""
import sys, os
from io import StringIO
import importlib
import traceback

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

class Capturing(list):
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self
    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio    # free up some memory
        sys.stdout = self._stdout

# set working path relative to this file
abspath = os.path.abspath(__file__)
dname = os.path.dirname(abspath)
os.chdir(dname)

# set environmental variables
os.environ["DIR_TESTS"] = "../../tests"
os.environ["DIR_SOURCE"] = "../../src"
os.environ["DIR_DOCS"] = "../../docs"
os.environ["DIR_TMP"] = "../../tmp"

def build():
    runonly = {}
    for path in sys.argv[1:]:
        if "::" in path:
            module, method = path.split("::")
        else:
            module, method = path, None
        if module not in runonly:
            runonly[module] = []
        if method:
            runonly[module].append(method)
            
    # run generator module
    for file_name in os.listdir():
        if file_name.startswith("build_"):
            module_name = file_name.replace(".py","")
            # run only selected modules
            if runonly and module_name not in runonly:
                continue
            module = importlib.import_module(module_name) 
            # run generator method
            print(f"{bcolors.HEADER}------ {module_name} ------{bcolors.ENDC}")
            for generator_name in dir(module):
                if generator_name.startswith("build_"):
                    # run only selected generators
                    if runonly and runonly[module_name] and generator_name not in runonly[module_name]:
                        continue
                    # run generator
                    try:
                        with Capturing() as output:
                            getattr(module,generator_name)()
                        if output:
                            output = "\n".join(output)
                            print(f"{bcolors.OKGREEN}{module_name}::{generator_name}{bcolors.ENDC}\n{output}")
                        else:
                            print(f"{bcolors.OKGREEN}{module_name}::{generator_name}{bcolors.ENDC}")
                    except Exception as error:
                        print(f"{bcolors.FAIL}{module_name}::{generator_name}{bcolors.ENDC}")
                        traceback.print_exc()
                        if output:
                            output = "\n".join(output)
                            print("\n"+output)
                            print(f"{bcolors.FAIL}{module_name}::{generator_name}{bcolors.ENDC}")
            print("")

if __name__ == "__main__":
    build()


import numpy as np
import os
import sys
sys.path.insert(0, os.environ['DIR_SOURCE'])

def build_thumbnail_image():
    
    from scinumtools import ThumbnailImage
    import matplotlib.pyplot as plt
    import numpy as np
    
    size = np.pi
    size1 = (100, 100)
    xaxis = np.linspace(-size, size, size1[0])
    yaxis = np.linspace(-size, size, size1[1])
    data = size*np.vectorize(lambda x,y: np.sin(x)*np.sin(y))(*np.meshgrid(xaxis,yaxis))
    extent1 = (xaxis.min(), xaxis.max(), yaxis.min(), yaxis.max())

    fig, axes = plt.subplots(2,2,figsize=(5,5),tight_layout=True)
    
    ticks1 = [-size,0,size]
    ticks2 = [-size*2,-size,0,size,size*2]
    labels1 = ['$-\pi$','0','$\pi$']
    labels2 = ['$-2\pi$','$-\pi$','0','$\pi$','$2\pi$']
    extent2 = (-size*2, size*2, -size*2, size*2)
    size2 = (20,20)
    plots = [
        ('a.', 0, 0, ticks1, labels1, ThumbnailImage(data, extent1, mode='F') ),
        ('b.', 1, 0, ticks1, labels1, ThumbnailImage(data, extent1).resize(size2) ),
        ('c.', 0, 1, ticks2, labels2, ThumbnailImage(data, extent1).crop(extent2) ),
        ('d.', 1, 1, ticks2, labels2, ThumbnailImage(data, extent1).crop(extent2, bgcolor=2).resize(size2) ),
    ]
    for name, m, n, ticks, labels, ti in plots:
        ax = axes[m,n]
        ti.draw(ax)
        ax.text(0.05, 0.9, name, transform=ax.transAxes)
        ax.text(0.05, 0.05, ti.im.size, transform=ax.transAxes)
        ax.set_xticks(ticks)
        ax.set_xticklabels(labels)
        ax.set_yticks(ticks)
        ax.set_yticklabels(labels)

    dir_figures = os.environ['DIR_DOCS']+"/source/_static/figures"
    file_figure = dir_figures+"/thumbnail_image.png"
    plt.savefig(file_figure)
    print(file_figure)

def build_normalize_data():
    
    from scinumtools import NormalizeData
    import matplotlib.pyplot as plt
    import numpy as np
    
    with NormalizeData(xaxis='lin', yaxis='lin') as nd:
        
        for size in [np.pi*2, np.pi]:
            xaxis = np.linspace(0,size,50)
            yaxis = np.linspace(-size,size,50)
            zdata = size*np.vectorize(lambda x,y: np.sin(x)*np.sin(y))(*np.meshgrid(xaxis,yaxis))
            nd.append(zdata, xaxis, yaxis)
        
        xranges = nd.xranges()
        yranges = nd.yranges()
        norm = nd.linnorm()
    
        fig, axes = plt.subplots(1,2,figsize=(5,3))
    
        for i, (data, extent) in enumerate(nd.items()):
            ax = axes[i]
            im = ax.imshow(data, extent=extent, norm=norm)
            ax.set_xlim(xranges.min, xranges.max)
            ax.set_ylim(yranges.min, yranges.max)
        
        fig.colorbar(im, ax=axes.ravel().tolist())
        
        dir_figures = os.environ['DIR_DOCS']+"/source/_static/figures"
        file_figure = dir_figures+"/normalize_data.png"
        plt.savefig(file_figure)
        print(file_figure)
    
def build_data_plot_grid():
    
    from scinumtools import DataPlotGrid
    import matplotlib.pyplot as plt
    
    dpg = DataPlotGrid(['a','b','c','d','e'],ncols=3,axsize=(1,1))
    
    fig, axes = plt.subplots(dpg.nrows, dpg.ncols, figsize=dpg.figsize, tight_layout=True)
    for i, m, n, v in dpg.items():
        ax = axes[m,n]
        ax.text(0.5, 0.5, v)
    for i, m, n in dpg.items(missing=True):
        ax = axes[m,n]
        ax.set_axis_off()
    dir_figures = os.environ['DIR_DOCS']+"/source/_static/figures"
    file_figure = dir_figures+"/data_plot_grid1.png"
    plt.savefig(file_figure)
    print(file_figure)
    
    fig.clf()
    
    fig, axes = plt.subplots(dpg.nrows, dpg.ncols, figsize=dpg.figsize, tight_layout=True)
    for i, m, n, v in dpg.items(transpose=True):
        ax = axes[m,n]
        ax.text(0.5, 0.5, v)
    for i, m, n in dpg.items(transpose=True, missing=True):
        ax = axes[m,n]
        ax.set_axis_off()
        
    dir_figures = os.environ['DIR_DOCS']+"/source/_static/figures"
    file_figure = dir_figures+"/data_plot_grid2.png"
    plt.savefig(file_figure)
    print(file_figure)



import sys, os
import pytest
import numpy as np
sys.path.insert(0, 'src')

from scinumtools.dip import DIP
from scinumtools.dip.docs import ExportDocsPDF
from scinumtools.dip.docs.settings import DocsType
from scinumtools.dip.docs.item_parameter import ParType

@pytest.fixture
def file_pdf():
    dir_tmp = "tmp" #docs/source/_static/pdf"
    if not os.path.isdir(dir_tmp):
        os.mkdir(dir_tmp)
    file_pdf  = f"{dir_tmp}/documentation.pdf"
    if os.path.isfile(file_pdf):
        os.remove(file_pdf)
    return file_pdf

@pytest.fixture
def file_definitions():
    return "../../docs/source/_static/pdf/definitions.dip"
    
@pytest.fixture
def file_cells():
    return "../../docs/source/_static/pdf/cells.dip"

def test_export_pdf(file_pdf, file_definitions, file_cells):
    with DIP() as p:
        p.add_unit("velocity", 13, 'cm/s')
        p.add_string("""
        $unit length = 1 cm
        $unit mass = 2 g
        
        cfl_factor float = 0.7  # Courant–Friedrichs–Lewy condition
        max_vare float = 0.2    # maximum energy change of electrons
        max_vari float = 0.2    # maximum energy change of ions
        """)
        p.add_source("cells", file_cells)
        p.add_file(file_definitions)
        docs = p.parse_docs()
    with ExportDocsPDF(docs) as exp:
        title = "Example documentation"
        pageinfo = "DIP Documentation"
        exp.build(file_pdf, title, pageinfo)


def check_node(node, name, keyword, value):
    assert node.name == name
    assert node.keyword == keyword
    assert node.value.value == value

def test_definition_before_case():
    
    with DIP(docs=True) as p:
        p.add_string("""
        a int = 3        # definition
        @case false
          a float = 4.0  # modification
        @case true
          a str = "4.0"  # modification
        @else
          a bool = true  # modification
        a = 4            # modification
        """)
        docs = p.parse_docs()
        assert len(docs.env.nodes) == 5
        check_node(docs.env.nodes[0], 'a', 'int', 3)
        
        assert 'a' in docs.parameters
        nodes = docs.parameters['a'].nodes
        assert nodes[0].ntype == ParType.DEF.value
        assert nodes[1].ntype == ParType.MOD.value
        assert nodes[2].ntype == ParType.MOD.value
        assert nodes[3].ntype == ParType.MOD.value
        assert nodes[4].ntype == ParType.MOD.value
    
def test_branch_definition():

    # full definition in branch
    with DIP(docs=True) as p:
        p.add_string("""
        @case false
          a float = 4.0  # definition
        @case true
          a str = "4.0"  # definition
        @else
          a bool = true  # definition
        a = 4            # modification
        """)
        docs = p.parse_docs()
        assert len(docs.env.nodes) == 4
        check_node(docs.env.nodes[0], '@1.a', 'float', 4.0)
        check_node(docs.env.nodes[1], '@2.a', 'str',   "4.0")
        check_node(docs.env.nodes[2], '@3.a', 'bool',  True)
        
        assert 'a' in docs.parameters
        nodes = docs.parameters['a'].nodes
        assert nodes[0].ntype == ParType.DEF.value
        assert nodes[1].ntype == ParType.DEF.value
        assert nodes[2].ntype == ParType.DEF.value
        assert nodes[3].ntype == ParType.MOD.value

def test_incomplete_case_definition():

    # definition in else case is missing
    with DIP(docs=True) as p:
        p.add_string("""
        @case true
          a float = 4.0  # definition
        a float = 3      # definition, or modification
        """)
        docs = p.parse_docs()
        assert len(docs.env.nodes) == 2
        check_node(docs.env.nodes[0], '@1.a', 'float', 4.0)
        check_node(docs.env.nodes[1], 'a',    'float', 3)

    # all but one branch has definition
    with DIP(docs=True) as p:
        p.add_string("""
        @case true
          b float = 4.0  
        @else
          a float = 5.0  # definition
        a float = 3      # modification
        """)
        docs = p.parse_docs()
        assert len(docs.env.nodes) == 3
        check_node(docs.env.nodes[0], '@1.b', 'float', 4.0)
        check_node(docs.env.nodes[1], '@2.a', 'float', 5.0)
        check_node(docs.env.nodes[2], 'a',    'float', 3.0)

def test_nested_incomplete():

    # incomplete nested definition
    with DIP(docs=True) as p:
        p.add_string("""
        @case true
          @case true
            b float = 4.0  
          @else
            a float = 8.0  # definition
        @else
          a float = 5.0  # definition
        a float = 3      # modification
        """)
        docs = p.parse_docs()
        assert len(docs.env.nodes) == 4
        check_node(docs.env.nodes[0], '@1.@2.b', 'float', 4.0)
        check_node(docs.env.nodes[1], '@1.@3.a', 'float', 8.0)
        check_node(docs.env.nodes[2], '@4.a',    'float', 5.0)
        check_node(docs.env.nodes[3], 'a',       'float', 3.0)



from bs4 import BeautifulSoup
import json
from pathlib import Path
from pygments import highlight
from pygments.lexers import PythonLexer
from pygments.formatters import get_formatter_by_name
from pygments.styles import get_style_by_name
from pygments.token import STANDARD_TYPES, Token
from pygments.formatters import HtmlFormatter
import re

from .settings import *
from ..item_parameter import ParType
from ..documentation import Documentation
from ...settings import Sign, ROOT_SOURCE
from ...pygments import SyntaxLexer, StyleLexer

class SettingsSection:
        
    docs: Documentation
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
    
    def __init__(self, docs: Documentation, **kwargs):
        self.docs = docs
        self.html = BeautifulSoup(f"", 'html.parser')
                
    def highlight_python_code(self, code, target):
        # format text using pygments
        lexer = SyntaxLexer()
        style = StyleLexer

        # format text with a lexer
        formatter = get_formatter_by_name(
            'html', style=style, cssclass='pt-2 pb-2',
            linenos='inline', wrapcode=True, 
            prestyles='margin-bottom: 0',
            lineanchors=target
        ) 
        code = highlight(code, lexer, formatter)
        code = code.replace(f"{target}-",f"{target}_")
        #code = code.replace(f"linenos",f"")
        return code
        
    def styles(self):
        style = StyleLexer
        formatter = HtmlFormatter(style=style)
        css = formatter.get_style_defs()
        css += "\n .linenos {width:50px; display: inline-block; text-align:right; margin-right: 10px; border-right: 1px grey solid; user-select: none;}"
        return css
        
    def build_units(self):
        section = BeautifulSoup(Title("Units",3), 'html.parser')
        self.html.append(section)
        
        table = self.html.new_tag('table', **{'class':'table table-bordered'})
        row = self.html.new_tag('tr', **{'class':'thead-light'})
        names = ['Name','Value','Units','Source']
        for name in names:
            col = self.html.new_tag('th', **{'class':'p-1 bg-body-secondary'})
            col.string = name
            row.append(col)
        table.append(row)
        for unit in self.docs.units:
            row = self.html.new_tag('tr', **{'class':''})
            values = [unit.name, unit.value, unit.units]
            for value in values:
                col = self.html.new_tag('td', **{'class':'p-1'})
                col.string = value
                row.append(col)
            col = self.html.new_tag('td', **{'class':'p-1'})
            col.append(Link(PAGE_SOURCES,unit.link_source,f"{unit.source[0]}:{unit.source[1]}"))
            row.append(col)
            table.append(row)
        self.html.append(table)
        
    def build_sources(self):
        section = BeautifulSoup(Title("Sources",3), 'html.parser')
        self.html.append(section)
        
        for sdata in self.docs.sources:
            source = self.html.new_tag("div", **{'class':"container mt-2 border bg-white"})
            
            header = self.html.new_tag('div', **{'class':'row bg-dark-subtle'})
            name = self.html.new_tag('div', **{'class':'col'})
            name.append(Target(sdata.target, sdata.name))
            header.append(name)
            source.append(header)
            
            props = self.html.new_tag('div', **{'class':'row'})
            name = self.html.new_tag('div', **{'class':'col-2 bg-body-secondary'})
            name.string = 'File:'
            props.append(name)
            value = self.html.new_tag('div', **{'class':'col bg-white'})
            value.string = Path(sdata.path).name
            props.append(value)
            source.append(props)

            if sdata.parent:
                props = self.html.new_tag('div', **{'class':'row'})
                name = self.html.new_tag('div', **{'class':'col-2 bg-body-secondary'})
                name.string = 'Source:'
                props.append(name)
                value = self.html.new_tag('div', **{'class':'col bg-white'})
                value.append(Link(PAGE_SOURCES, sdata.link_source, f"{sdata.parent[0]}:{sdata.parent[1]}"))
                props.append(value)
                source.append(props)
            
            if sdata.code:
                code = self.highlight_python_code(sdata.code, sdata.target)
                code = BeautifulSoup(code, 'html.parser') 
                source.append(code)

            self.html.append(source)
        
    def build(self):
        
        self.build_units()
        
        self.build_sources()

        return self.html


import os
import numpy as np
import hashlib
import json

def hash_file_name(file_cache, args, kwargs):
    json_string = json.dumps({'args':args,'kwargs':kwargs}, sort_keys=True)
    hash_object = hashlib.sha256()
    hash_object.update(json_string.encode())
    hash_value = hash_object.hexdigest()
    file_name, file_extension = os.path.splitext(file_cache)
    return f"{file_name}.{hash_value}{file_extension}"

def CachedFunction(file_cache):
    """ Decorator function that cashes data produced by a function

    :param str file_cache: Name of the file that will store cached data
    """
    def wrapped(func):
        def inner(*args, **kwargs):
            file_data = hash_file_name(file_cache, args, kwargs)
            if not os.path.isfile(file_data):
                np.save(file_data, func(*args, **kwargs))
            data = np.load(file_data, allow_pickle=True)
            return data.item() if data.ndim==0 else data
        return inner
    return wrapped



from reportlab.platypus import Table, Paragraph, Spacer, PageBreak
from reportlab.lib.units import inch
import numpy as np

from .settings import *
from ...settings import ROOT_SOURCE

class UnitsSection:
    
    data: list
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
        
    def __init__(self, data: list):
        self.data = data
        
    def parse_table(self):
        TABLE_STYLE = [
            ('GRID',       (0,1), (-1,-1),  0.5, PALETTE['prop_name']),
            ('BACKGROUND', (0,0), (-1,0),   PALETTE['node_name']),  
            ('BACKGROUND', (0,1), (0,-1),   PALETTE['prop_name']),  
            ('BACKGROUND', (1,1), (-1,-1),    PALETTE['prop_value']),  
        ]
    
        data = [   
            ['Name','Value','Units','Source'],
        ]
        for item in self.data:
            src = Paragraph(Link(item.link_source, f"{item.source[0]}:{item.source[1]}")) 
            # add row to the table
            data.append([item.name, item.value, item.units, src])
            
        colWidths = list(np.array([0.2,0.2,0.2,0.4])*(PAGE_WIDTH-2*inch))
        return Table(data, style=TABLE_STYLE, hAlign='LEFT', colWidths=colWidths)
        
    def parse(self):
        blocks = []
        blocks.append(Paragraph(Title(f"List of units"), H2))
        blocks.append(self.parse_table()) 
        blocks.append(Spacer(1,0.2*inch))
        return blocks
    


import time
from textwrap import dedent
import sys
sys.path.insert(0, 'src')

from scinumtools import Stopwatch

def test_normal_object():
    # Using as a normal object
    sw = Stopwatch()
    sw.start('part1')
    time.sleep(.01)
    sw.start('part2')
    time.sleep(.02)
    sw.stop('part2')
    sw.stop('part1')
    sw.start('part3')
    time.sleep(.05)
    sw.stop('part3')
    result = sw.report().to_dict()
    assert result['Laps'] == [6, 1, 1, 1]
    assert result['Node'] == ['stopwatch', 'part1/part2', 'part1', 'part3']

def test_with_statement():
    with Stopwatch() as sw:
        with sw.observer('part1'):
            time.sleep(.01)
            with sw.observer('part2'):
                time.sleep(.02)
        with sw.observer('part3'):
            time.sleep(.05)
        result = sw.report().to_dict()
        assert result['Laps'] == [6, 1, 1, 1]
        assert result['Node'] == ['stopwatch', 'part1/part2', 'part1', 'part3']



import numpy as np

from .settings import *
from ..nodes import ModNode, ImportNode, StringNode
from ..settings import Sign, Keyword, EnvType

class NodeItem:
    target: str
    link_source: str
    link_injection: str
    link_import: str
    name: str
    value: str
    unit: str
    dtype: str
    source: tuple
    unsigned: bool
    precision: int
    injection: bool
    imported: tuple
    options: list
    constant: bool
    dformat: str
    tags: list
    description: str
    condition: str
               
    def __init__(self, name, node, node_type):
        self.target = NodeItemTarget(node.name, node.source[0], node.source[1])
        self.link_source = SourceItemTarget(node.source[0], node.source[1])
        self.link_injection = InjectionItemTarget(node.name, node.source[0], node.source[1])
        self.name = name
        self.source = node.source
        self.ntype = node_type.value
        # type specific initialization
        self.value = None
        self.options = None
        self.unsigned = None
        self.precision = None
        getattr(self,f"_init_{node.keyword}")(node)

        # find out units
        self.unit = ''
        if node.keyword != ModNode.keyword and node.value and node.value.unit:
            self.unit = node.value.unit
        elif node.units_raw:
            self.unit = node.units_raw
            
        # data type
        self.dtype = node.keyword
        if self.unsigned:
            self.dtype = f"u{self.dtype}"
        if self.precision:
            self.dtype = f"{self.dtype}{self.precision}"
        
        # references
        self.injection = True if node.value_ref else False
        self.imported = node.isource if isinstance(node.isource,tuple) else False
        if self.imported:
            self.link_import = ImportItemTarget(node.isource[0], node.isource[1])
        
        # node attributes
        self.constant = node.constant
        self.dformat = node.format if node.keyword == StringNode.keyword else None
        if node.keyword == ModNode.keyword or node.tags is None:
            self.tags = []
        else:
            self.tags = node.tags
        self.description = None if node.keyword == ModNode.keyword else node.description
        self.condition = None
        if node.condition:
            self.condition = node.condition
            
    def _init_bool(self, node):
        if node.value:
            self.value = Keyword.TRUE if node.value.value else Keyword.FALSE
            
    def _init_int(self, node):
        if node.value:
            if isinstance(node.value.value, (np.ndarray, list)):
                self.value = str(node.value.value)
            else:
                self.value = str(int(node.value.value))
        if node.units_raw:
            self.options = [f"{option.value.value} {option.value.unit}" for option in node.options]
        else:
            self.options = [f"{option.value.value}" for option in node.options]
        self.unsigned = node.unsigned
        self.precision = node.precision

    def _init_float(self, node):
        if node.value:
            if isinstance(node.value.value, (np.ndarray, list)):
                self.value = str(node.value.value)
            else:
                exp = np.log10(node.value.value)
                if exp<=3 or exp>=-3:
                    self.value = f"{node.value.value:.03f}"
                else:
                    self.value = f"{node.value.value:.03e}"
        if node.units_raw:
            self.options = [f"{option.value.value} {option.value.unit}" for option in node.options]
        else:
            self.options = [f"{option.value.value}" for option in node.options]
        self.precision = node.precision
        
    def _init_str(self, node):
        if node.value:
            self.value = str(node.value.value)
        self.options = [str(option.value.value) for option in node.options]
            
    def _init_mod(self, node):
        self.value = node.value



import numpy as np
from collections import deque
import time

class ProgressBar:
    current: int = 0    # current step
    nsteps: int         # number of steps
    times: deque        # list of step times
    ntimes: int = 20    # number of times for a sliding mean
    time_start: float   # starting time
    size: int = 70      # size of the progress bar
    
    def __enter__(self):
        return self
        
    def __exit__(self, type, value, traceback):
        self.close()
        
    def __init__(self, nsteps: int):
        self.nsteps = nsteps
        self.time_start = time.time()
        self.times = deque([self.time_start])
        print(f"|{self._info_text():{self.size}s}|")
        
    def _time_text(self, seconds):
        if seconds*2.777777e-4>=1:
            return f"{seconds*2.777777e-4:.1f}h"
        elif seconds*1.666666e-2>=1:
            return f"{seconds*1.666666e-2:.1f}m"
        else:
            return f"{seconds:.1f}s"
            
    def _info_text(self):
        text =  f"Step {self.current:d}/{self.nsteps:d}"
        if len(self.times)>1:
            times = np.array(self.times)
            dt = times[1:]-times[:-1]
            dt_mean = np.sum(dt)/len(dt)
            rt = dt_mean * (self.nsteps-self.current)
        else:
            rt = 0
        et = time.time()-self.time_start
        text += f" Time "+self._time_text(et)
        text += f"/"+self._time_text(rt+et)
        return text
        
    def close(self):
        self.current = self.nsteps-1
        self.step()
        
    def step(self, info: int = ''):
        self.current += 1
        # Get progress bar text
        pbar = self._info_text()
        if info:
            pbar += f"   {info}"
        # Clean previous line
        print("\033[A"+" "*(self.size+2)+f"\033[A")
        # Print a new line
        i = int(np.floor(self.size*self.current/self.nsteps))
        pbar = f"{pbar:{self.size}s}"
        pbar = "\033[;7m"+pbar[:i]+"\033[;0m"+pbar[i:]+"\033[;0m";
        print(f"|{pbar}|")
        # Update deque
        self.times.append(time.time())
        if len(self.times)>self.ntimes:
            self.times.popleft()



from .list_nodes import NodeList
from .list_sources import SourceList
from .list_units import UnitList
from .list_hierarchy import HierarchyList
from .list_branching import BranchingList
from .list_functions import FunctionList


from reportlab.platypus import Paragraph, Table, Spacer, PageBreak
from reportlab.lib.units import inch
import numpy as np
import re

from .settings import *
from ..settings import DocsType
from ...nodes import Node, ImportNode
from ...settings import Order, Sign, Keyword, EnvType
from ...environment import Environment

class ImportsSection:
    
    data: list       # list of imports
    
    def __enter__(self):
        return self
    def __exit__(self, type, value, traceback):
        pass
        
    def __init__(self, data):
        self.data = data

    def parse_item(self, item):
    
        # define table style
        TABLE_STYLE = [
            # whole grid
            ('GRID',       (0,1), (-1,-1),  0.5, PALETTE['prop_name']),
            ('FONTNAME',   (0,0), (-1,-1),  TABLE_BODY_STYLE.fontName),
            ('FONTSIZE',   (0,0), (-1,-1),  TABLE_BODY_STYLE.fontSize),
            # top panel
            ('FONTSIZE',   (0,0), (-1,-1),  TABLE_HEADER_STYLE.fontSize),
            ('TEXTCOLOR',  (0,0), (-1,0),   PALETTE['c4']),   
            ('BACKGROUND', (0,0), (-1,0),   PALETTE['node_name']),
            ('ALIGN',      (-2,0),(-1,0),   'RIGHT'),
            ('BACKGROUND', (0,1), (1,1),   PALETTE['prop_name']),        # request title
            ('BACKGROUND', (0,2), (2,2),   PALETTE['prop_name']),        # node name title
            ('BACKGROUND', (1,1), (-1,1),  PALETTE['prop_value']),      # request value
            ('SPAN',       (0,0), (2,0)     ),                            # parameter name
            ('SPAN',       (1,1), (2,1)     ),                            # request
            ('SPAN',       (0,2), (1,2)    ),                            # original names
        ]
        
        # construct a node table
        target_import = Target(item.target) 
        link_source = Link(item.link_source, f"{item.source[0]}:{item.source[1]}") 
        data = [
            [Paragraph(target_import+link_source), '', ''],
        ]
        TABLE_STYLE.append(('SPAN', (2,len(data)), (-1,len(data)) ))
        data.append(['Request:', Paragraph(HighlightReference("{"+item.reference+"}"), TABLE_BODY_STYLE)])
        rows = []
        for idata in item.idata:
            nrow = len(data)+1+len(rows)
            TABLE_STYLE.append(('SPAN', (0,nrow), (1,nrow) ))
            rows.append([
                Paragraph(Link(idata.link_node, idata.name), TABLE_BODY_STYLE),  
                '',
                Paragraph(Link(idata.link_source, f"{idata.source[0]}:{idata.source[1]}"), TABLE_BODY_STYLE),  
            ])
        if rows:
            TABLE_STYLE.append(('BACKGROUND', (0,3), (-1,-1),  PALETTE['prop_value']))
            data.append(['Imported node:', '',  'From source:'])
            data += rows
                        
        return Table(data,style=TABLE_STYLE, hAlign='LEFT', colWidths=ColumnWidths([0.2, 0.5, 0.3]))
        
    def parse(self):
        blocks = []
        blocks.append(Paragraph(Title(f"Imported nodes"), H2) )
        for item in self.data:
            blocks.append(Spacer(1,0.1*inch))
            blocks.append(self.parse_item(item))
        blocks.append(Spacer(1,0.2*inch))
        return blocks


from enum import Enum, auto

from .settings import *
from .item_node import NodeItem

class ParType(Enum):
    DEC    = 0    # declaration
    DEF    = 1    # definition
    DCM    = 2    # declaration or modification
    DFM    = 3    # definition or modification
    MOD    = 4    # modification
    INJ    = 5    # injection
    IMP    = 6    # import
    
class ParameterItem:
    target: str
    name: str
    counts: list
    nodes: list
        
    def __init__(self, name, node):
        self.target = ParameterItemTarget(name)
        self.name    = name
        self.counts  = [0]*len(ParType)
        self.nodes   = []
        self.add(name, node)
        
    def add(self, name, node):
        # update node type counters
        if DocsType.DECLARATION|DocsType.MODIFICATION in node.docs_type:
            node_type = ParType.DCM
        elif DocsType.DEFINITION|DocsType.MODIFICATION in node.docs_type:
            node_type = ParType.DFM
        elif DocsType.DECLARATION in node.docs_type:
            node_type = ParType.DEC
        elif DocsType.DEFINITION in node.docs_type:
            node_type = ParType.DEF
        elif DocsType.MODIFICATION in node.docs_type:
            node_type = ParType.MOD
        self.counts[node_type.value] += 1
        # update reference counters
        if node.value_ref:
            self.counts[ParType.INJ.value] += 1
        if node.isource:
            self.counts[ParType.IMP.value] += 1
        # add node to the list
        self.nodes.append(NodeItem(name, node, node_type))



import numpy as np

from .export_c import ExportConfigC
from ..datatypes import StringType, BooleanType, NumberType, FloatType, IntegerType
from ..settings import Sign

class ExportConfigCPP(ExportConfigC):
    
    def _parse_dtype(self, param):
        if isinstance(param, StringType):
            dtype = "char*"
        elif isinstance(param, BooleanType):
            dtype = "bool"
        elif isinstance(param, IntegerType):
            if param.unsigned:
                if param.precision==16:   dtype = "unsigned short int"
                elif param.precision==32: dtype = "unsigned int"
                elif param.precision==64: dtype = "unsigned long long int"
            else:
                if param.precision==8:    dtype = "signed char"
                elif param.precision==16: dtype = "short int"
                elif param.precision==32: dtype = "int"
                elif param.precision==64: dtype = "long long int"
        elif isinstance(param, FloatType):
            if param.precision==32: dtype = "float"
            elif param.precision==64: dtype = "double"
            elif param.precision in [80,96,128]: dtype = "long double"
        return dtype

    def parse_constexpr(self, name, param):
        name = self._rename(name)
        dtype = self._parse_dtype(param)
        value, shape = self._parse_value(param)
        if shape is None:
            return f"constexpr {dtype} {name} = {value};"
        else:
            shape = "[" + "][".join(str(s) for s in shape) + "]"
            return f"constexpr {dtype} {name}{shape} = {value};"
            
    def parse(self, guard:str="CONFIG_H", define:tuple=None, const:tuple=None):
        lines = []
        lines.append(f"#ifndef {guard}")
        lines.append(f"#define {guard}")
        lines.append("")
        for name, param in self.data.items():
            if define and name in define:
                lines.append(self.parse_define(name, param))
            elif const and name in const:
                lines.append(self.parse_const(name, param))
            else:
                lines.append(self.parse_constexpr(name, param))
        lines.append("")
        lines.append(f"#endif /* {guard} */")
        if self.includes:
            lines.insert(3,"")
            for inc in self.includes:
                lines.insert(3,f"#include {inc}")
        self.text = Sign.NEWLINE.join(lines)
        return self.text
    


from reportlab.platypus import Table, Paragraph, Spacer, PageBreak
from reportlab.lib.units import inch
import numpy as np

from .settings import *

class TypesSection:
    
    data: dict
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
        
    def __init__(self, data: dict):
        self.data = data
        
    def parse_table(self):
        
        TABLE_STYLE = [
            ('BACKGROUND', (0,0), (0,0),   NTYPES[self.data[0]][1]),  
            ('BACKGROUND', (0,1), (0,1),   NTYPES[self.data[1]][1]),  
            ('BACKGROUND', (0,2), (0,2),   NTYPES[self.data[2]][1]),  
            ('BACKGROUND', (0,3), (0,3),   NTYPES[self.data[3]][1]),  
            ('BACKGROUND', (0,4), (0,4),   NTYPES[self.data[4]][1]),  
            ('BACKGROUND', (2,0), (2,0),   NTYPES[self.data[5]][1]),  
            ('BACKGROUND', (2,1), (2,1),   NTYPES[self.data[6]][1]),  
        ]
    
        data = [   
            ['',NTYPES[self.data[0]][0],'',NTYPES[self.data[5]][0] ],
            ['',NTYPES[self.data[1]][0],'',NTYPES[self.data[6]][0] ],
            ['',NTYPES[self.data[2]][0] ],
            ['',NTYPES[self.data[3]][0] ],
            ['',NTYPES[self.data[4]][0] ],
        ]
        return Table(data, style=TABLE_STYLE, hAlign='LEFT', colWidths=ColumnWidths([0.05,0.45,0.05,0.45]))
        
    def parse(self):
        blocks = []
        blocks.append(Paragraph(Title(f"Node types"), H2))
        blocks.append(self.parse_table()) 
        blocks.append(Spacer(1,0.2*inch))
        return blocks
    


from .parser import Parser
from .node import Node
from .node_empty import EmptyNode
from .node_import import ImportNode
from .node_unit import UnitNode
from .node_source import SourceNode
from .node_case import CaseNode
from .node_option import OptionNode
from .node_constant import ConstantNode
from .node_format import FormatNode
from .node_condition import ConditionNode
from .node_group import GroupNode
from .node_import import ImportNode
from .node_mod import ModNode
from .node_boolean import BooleanNode
from .node_integer import IntegerNode
from .node_float import FloatNode
from .node_string import StringNode
from .node_table import TableNode
from .node_tags import TagsNode
from .node_description import DescriptionNode



import numpy as np
import os, sys
sys.path.insert(0, os.environ['DIR_SOURCE'])
import requests
from bs4 import BeautifulSoup
import json
import re

from scinumtools import ProgressBar, CachedFunction
from scinumtools.units import Quantity

path_materials = os.environ['DIR_SOURCE']+'/scinumtools/materials'

def build_periodic_table():
    
    @CachedFunction(os.environ['DIR_TMP']+'/element.npy')
    def load_elements(url):
        response = requests.get(url)
        if response.status_code == 200:
            return response.text
        else:
            return None
            
    html_data = load_elements("https://physics.nist.gov/cgi-bin/Compositions/stand_alone.pl")
    if html_data:
        
        text = [
            "################################################################",
            "# Do not modify this file!                                     #",
            "# It is generated automatically in:                            #",
            "# tools/generator/build_materials.py                           #",
            "# Isotope data was taken from NIST:                            #",
            "# https://physics.nist.gov/cgi-bin/Compositions/stand_alone.pl #",
            "################################################################",
            "",
            "PT_HEADER = ['Z','A']",
            "PT_DATA = {",
        ]
        
        soup = BeautifulSoup(html_data, "html.parser")
        table = soup.find("table")
        rows = table.find_all("tr")
        new = None
        Z, A, M, NA, symbol = 0, 0, 0, 0, None
        for row in rows:
            dts = row.find_all('td',{'align':'right'})
            if len(dts)==1:
                A = int(dts[0].get_text().strip())
                new = False
            elif len(dts)==2:
                Z = int(dts[0].get_text().strip())
                A = int(dts[1].get_text().strip())
                symbol = row.find('td',{'align':'center'}).get_text().strip()
                if new is not None:
                    text.append("    }),")
                text.append("    '"+symbol+"': ("+str(Z)+", {")
                new = True
            else:
                continue
            dts = row.find_all('td',{'align':None})
            # relative atomic masses
            M = float(re.sub(r"(\s+|\([0-9#]+\))", '', dts[0].get_text()))
            # isotopic composition
            if ic := dts[1].get_text().strip():
                NA = float(re.sub(r"(\s+|\([0-9#]+\))", '', ic))
            else:
                NA = 0
            text.append(f"        '{A}': ({M}, {NA}),")
            
        text += [
            "    }),",
            "}", 
        ]
        text = "\n".join(text)
        
        # test if we produced a valid Python code
        exec(text, globals())
        assert PT_HEADER
        assert PT_DATA
        
        print('hello')
        
        # save the new version of the code
        path_units = os.environ['DIR_SOURCE']+"/scinumtools/materials"
        path_list = f'{path_units}/periodic_table.py'
        assert os.path.isdir(path_units)
        with open(path_list,'w') as f:
            f.write(text)
        print(path_list)


import numpy as np
import pytest
from math import isclose
import os
import sys
from decimal import Decimal
sys.path.insert(0, 'src')

from scinumtools import ParameterTable
from scinumtools.units import *

def test_quantity():
    
    assert str(Quantity(123e2))          == "Quantity(1.230e+04)"
    q = Quantity(123e2, 'km/s')
    assert str(q) == "Quantity(1.230e+04 km*s-1)"
    assert str(q.value())                == "12300.0"
    assert str(q.value('m/s'))           == "12300000.0"
    assert str(q.value('m/s',dtype=int)) == "12300000"
    assert str(q.value({'m':1,'s':-1}))  == "12300000.0"
    assert str(q.units())                == "km*s-1"
    
    result = "Quantity(1.230e+04 m*s2:3)"
    assert str(Quantity(123e2, [1,0,(2,3),0,0,0,0,0] ))    == result
    assert str(Quantity(123e2, Dimensions(m=Fraction(1), s=Fraction(2,3)) )) == result

    result = "Quantity(1.230e+04 J2*kg2:3)"
    assert str(Quantity(123e2, {'J': 2, 'k:g':(2,3)} )) == result
    assert str(Quantity(123e2, BaseUnits({'J': 2, 'k:g':(2,3)}) )) == result

    result = "Quantity(2.460e+04 m)"
    assert str(Quantity(123e2, Quantity(2, 'm'))) == result
    assert str(Quantity(123e2, 2*Unit('m'))) == result
    
def test_units():

    assert str(Unit('m'))                   == "Quantity(1.000e+00 m)"
    assert str(Unit('kg*m2*s-2'))           == "Quantity(1.000e+00 kg*m2*s-2)"
                                            
    unit = Unit()                           
    assert str(unit.m)                      == "Quantity(1.000e+00 m)"
    assert str(2*unit.kJ)                   == "Quantity(2.000e+00 kJ)"
    assert str(unit.kg*unit.m**2/unit.s**2) == "Quantity(1.000e+00 kg*m2*s-2)"
    assert str(unit.J.to('erg'))            == "Quantity(1.000e+07 erg)"
    assert str(unit.m**(2,3))               == "Quantity(1.000e+00 m2:3)"
    assert str(unit.m**Fraction(2,3))       == "Quantity(1.000e+00 m2:3)"
    assert str((9*unit.m)**(1,2))           == "Quantity(3.000e+00 m1:2)"

    with Unit() as unit:
        assert str(unit.m)                  == "Quantity(1.000e+00 m)"

def test_constants():

    assert str(Constant('c')) == "Quantity(1.000e+00 [c])"

    const = Constant()
    assert str(const.c)       == "Quantity(1.000e+00 [c])"
    assert str(const.m_e)     == "Quantity(1.000e+00 [m_e])"

    with Constant() as const:
        assert str(const.c)   == "Quantity(1.000e+00 [c])"
    
def test_aliases():
    assert str(quant(23, 'km'))   == "Quantity(2.300e+01 km)"
    assert str(unit('m'))         == "Quantity(1.000e+00 m)"
    assert str(const('c'))        == "Quantity(1.000e+00 [c])"    

def test_scalar_arithmetics():

    # addition and substtraction
    assert str(Quantity(1.1, 'km')-Quantity(100, 'm')) == "Quantity(1.000e+00 km)"
    assert str(Quantity(1.0, 'km')+Quantity(100, 'm')) == "Quantity(1.100e+00 km)"

    with pytest.raises(Exception) as e_info:
        Quantity(10, 'm') + Quantity(1, 'J')
    assert e_info.value.args[0] == "Unsupported addition between units:"

    with pytest.raises(Exception) as e_info:
        Quantity(10, 'm') - Quantity(1, 'J')
    assert e_info.value.args[0] == "Unsupported subtraction between units:"
    
    # multiplication, division and power
    q = Quantity(123e2, [3,3,0,0,1,0,0,0])
    assert str(q) == "Quantity(1.230e+04 m3*g3*C)"    
    q /= Quantity(123, 'C')
    assert str(q) == "Quantity(1.000e+02 m3*g3)"
    q *= Quantity(2, 's2')
    assert str(q) == "Quantity(2.000e+02 m3*g3*s2)"
    q = Quantity(123, "kg3*cm-2*s")
    assert str(q) == "Quantity(1.230e+02 kg3*cm-2*s)"
    q = Quantity(123e34, "J")
    assert str(q) == "Quantity(1.230e+36 J)"
    q /= Quantity(123, 's')
    assert str(q) == "Quantity(1.000e+34 J*s-1)"
    q = q ** 2
    assert str(q) == "Quantity(1.000e+68 J2*s-2)"
    q = q.to('kg2*m4/s6')
    assert str(q) == "Quantity(1.000e+68 kg2*m4*s-6)"
    q = Quantity(123, "kg3*s/(cm2*m)")
    assert str(q) == "Quantity(1.230e+02 kg3*s*cm-2*m-1)"
    q = q.to("kg3*s/cm3")
    assert str(q) == "Quantity(1.230e+00 kg3*s*cm-3)"
    q = Quantity(134e-34)
    assert str(q) == "Quantity(1.340e-32)"
    
    # arithmetics of dimensionless quantities
    q = Quantity(34)
    assert str(q-1) == "Quantity(3.300e+01)"

def test_scalar_unit_conversion():
    
    # error if converting different units
    q = Quantity(134e-34)
    with pytest.raises(Exception) as excinfo:
        q = q.to("kg3*s/cm3")
    assert excinfo.value.args[0]=="Unsupported conversion between units:"
    assert excinfo.value.args[1]==None
    assert excinfo.value.args[2]=='kg3*s*cm-3'

    # test unit conversion on scalars
    result = "Quantity(3.000e+03 m)"
    assert str(Quantity(3, 'km').to('m'))                 == result
    assert str(Quantity(3, 'km').to([1,0,0,0,0,0,0,0]))   == result
    assert str(Quantity(3, 'km').to(Dimensions(m=Fraction(1))))     == result
    assert str(Quantity(3, 'km').to({'m':1}))             == result
    assert str(Quantity(3, 'km').to(BaseUnits({'m':1})))  == result
    assert str(Quantity(3, 'km').to(Unit().m))            == result

    # reset base units if dimensions are all zero
    assert str(Quantity(3, 'kg*m2/s2')/Quantity(2, 'J')) == "Quantity(1.500e+00)"

    # reset base units, but take into account prefixes
    assert str(Quantity(10, 'm')/Quantity(5, 'cm'))      == "Quantity(2.000e+02)"

def test_operation_sides():
    
    p = Quantity(2, 'm')
    q = Quantity(3, 'm')
    assert p+q == q+p
    assert p-q == -(q-p)
    assert q*2 == 2*q
    assert p/2 == 1/(2/p)
    
def test_temperatures():
    
    # Temperature conversions
    assert str(Quantity(23,'K').to('Cel'))        == "Quantity(-2.501e+02 Cel)"
    assert str(Quantity(23,'K').to('degF'))       == "Quantity(-4.183e+02 degF)"
    assert str(Quantity(23,'K').to('degR'))       == "Quantity(4.140e+01 degR)"
    assert str(Quantity(23,'Cel').to('K'))        == "Quantity(2.961e+02 K)"
    assert str(Quantity(23,'Cel').to('degF'))     == "Quantity(7.340e+01 degF)"
    assert str(Quantity(23,'Cel').to('degR'))     == "Quantity(5.331e+02 degR)"
    assert str(Quantity(23,'degF').to('K'))       == "Quantity(2.681e+02 K)"
    assert str(Quantity(23,'degF').to('Cel'))     == "Quantity(-5.000e+00 Cel)"
    assert str(Quantity(23,'degF').to('degR'))    == "Quantity(4.827e+02 degR)"
    assert str(Quantity(23,'degR').to('K'))       == "Quantity(1.278e+01 K)"
    assert str(Quantity(23,'degR').to('Cel'))     == "Quantity(-2.604e+02 Cel)"
    assert str(Quantity(23,'degR').to('degF'))    == "Quantity(-4.367e+02 degF)"
    assert str(Quantity(2300,'Cel').to('kK'))     == "Quantity(2.573e+00 kK)"
    
def test_inversion():

    assert str(Quantity(23, 'Hz').to('s'))        == "Quantity(4.348e-02 s)"      # frequency to period
    assert str(Quantity(34, 'Ohm').to('S'))       == "Quantity(2.941e-02 S)"      # Ohm to Siemens
    assert str(Quantity(102, 'J').to('erg-1'))    == "Quantity(9.804e-10 erg-1)"

def test_nodim_to_radians():
    
    assert str(Quantity(23).to('rad'))            == "Quantity(2.300e+01 rad)"

def test_logarithmic_conversions():

    # Bels, Nepers and Amplitude/Power Ratios
    assert str(Quantity(1, 'B').to('dB'))     == "Quantity(1.000e+01 dB)"
    assert str(Quantity(1, 'dB').to('B'))     == "Quantity(1.000e-01 B)"
    assert str(Quantity(1, 'B').to('Np'))     == "Quantity(1.151e+00 Np)"
    assert str(Quantity(1, 'Np').to('B'))     == "Quantity(8.686e-01 B)"
    assert str(Quantity(1, 'dB').to('cNp'))   == "Quantity(1.151e+01 cNp)"
    
    assert str(Quantity(1000, 'AR').to('dB')) == "Quantity(6.000e+01 dB)"
    assert str(Quantity(30, 'dB').to('AR'))   == "Quantity(3.162e+01 AR)"
    assert str(Quantity(1000, 'PR').to('dB')) == "Quantity(3.000e+01 dB)"
    assert str(Quantity(6, 'dB').to('PR'))    == "Quantity(3.981e+00 PR)"

    assert str(Quantity(3.16228, 'AR').to('Np')) == "Quantity(1.151e+00 Np)"
    assert str(Quantity(1, 'Np').to('AR'))       == "Quantity(2.718e+00 AR)"
    assert str(Quantity(7.389, 'PR').to('Np'))   == "Quantity(1.000e+00 Np)"
    assert str(Quantity(0.115, 'Np').to('PR'))   == "Quantity(1.259e+00 PR)"

    # Decibel-milliwatts (dBm)
    assert str(Quantity(1, 'mW').to('W'))     == "Quantity(1.000e-03 W)"
    assert str(Quantity(1, 'mW').to('dBm'))   == "Quantity(0.000e+00 dBm)"
    assert str(Quantity(10, 'W').to('dBm'))   == "Quantity(4.000e+01 dBm)"
    assert str(Quantity(10, 'pW').to('dBm'))  == "Quantity(-8.000e+01 dBm)"
    assert str(Quantity(-3, 'dBm').to('uW'))  == "Quantity(5.012e+02 uW)"
    assert str(Quantity(22, 'dBm').to('mW'))  == "Quantity(1.585e+02 mW)"
    # Decibel-milliwatts (dBmW)
    assert str(Quantity(10, 'W').to('dBmW'))   == "Quantity(4.000e+01 dBmW)"
    assert str(Quantity(-3, 'dBmW').to('uW'))  == "Quantity(5.012e+02 uW)"
    # Decibel-watts (dBW)
    assert str(Quantity(1, 'W').to('dBW'))     == "Quantity(0.000e+00 dBW)"
    assert str(Quantity(20,'dBW').to('W'))     == "Quantity(1.000e+02 W)"
    # Conversions btw. dBm, dBmW and dBW
    assert str(Quantity(2, 'dBW').to('dBm'))    == "Quantity(3.200e+01 dBm)"
    assert str(Quantity(-20, 'dBm').to('dBW'))  == "Quantity(-5.000e+01 dBW)"
    assert str(Quantity(2, 'dBW').to('dBmW'))   == "Quantity(3.200e+01 dBmW)"
    assert str(Quantity(-20, 'dBmW').to('dBW')) == "Quantity(-5.000e+01 dBW)"
    assert str(Quantity(2, 'dBm').to('dBmW'))   == "Quantity(2.000e+00 dBmW)"
    assert str(Quantity(-20, 'dBmW').to('dBm')) == "Quantity(-2.000e+01 dBm)"
    
    # Decibel-Volt (dBV)
    assert str(Quantity(100, 'mV').to('dBV'))  == "Quantity(-2.000e+01 dBV)"
    assert str(Quantity(1, 'V').to('dBV'))     == "Quantity(0.000e+00 dBV)"
    assert str(Quantity(10, 'V').to('dBV'))    == "Quantity(2.000e+01 dBV)"
    assert str(Quantity(-60, 'dBV').to('mV'))  == "Quantity(1.000e+00 mV)"
    assert str(Quantity(0, 'dBV').to('V'))     == "Quantity(1.000e+00 V)"
    assert str(Quantity(40, 'dBV').to('V'))    == "Quantity(1.000e+02 V)"
    # Decibel-microvolt (dBuV)
    assert str(Quantity(42, 'dBuV').to('uV'))   == "Quantity(1.259e+02 uV)"
    assert str(Quantity(1000, 'uV').to('dBuV')) == "Quantity(6.000e+01 dBuV)"
    # Conversions btw. dBV and dBuV
    assert str(Quantity(10, 'dBV').to('dBuV'))   == "Quantity(1.300e+02 dBuV)"
    assert str(Quantity(-20, 'dBuV').to('dBV'))  == "Quantity(-1.400e+02 dBV)"
    
    # Decibel-microamps (dBA, dBuA)
    assert str(Quantity(10, 'dBA').to('A'))      == "Quantity(3.162e+00 A)"
    assert str(Quantity(0.1, 'A').to('dBA'))    == "Quantity(-2.000e+01 dBA)"
    assert str(Quantity(-40, 'dBuA').to('uA'))   == "Quantity(1.000e-02 uA)"
    assert str(Quantity(1000, 'uA').to('dBuA'))  == "Quantity(6.000e+01 dBuA)"
    
    # Decibel-ohm (dBOhm)
    assert str(Quantity(10, 'dBOhm').to('Ohm'))   == "Quantity(3.162e+00 Ohm)"
    assert str(Quantity(100, 'Ohm').to('dBOhm'))  == "Quantity(4.000e+01 dBOhm)"
    
    # Sound pressure level (dBSPL)
    assert str(Quantity(110, 'Pa').to('dBSPL'))  == "Quantity(1.348e+02 dBSPL)"
    assert str(Quantity(10, 'dBSPL').to('Pa'))   == "Quantity(6.325e-05 Pa)"
    # Sound intensity level (dBSIL)
    assert str(Quantity(100, 'W/m2').to('dBSIL'))  == "Quantity(1.400e+02 dBSIL)"
    assert str(Quantity(100, 'dBSIL').to('W/m2'))   == "Quantity(1.000e-02 W*m-2)"
    # Sound power level (dBSWL)
    assert str(Quantity(100, 'W').to('dBSWL'))  == "Quantity(1.400e+02 dBSWL)"
    assert str(Quantity(100, 'dBSWL').to('W'))   == "Quantity(1.000e-02 W)"
    
    # Units with for of: dBx/y
    q = Quantity(10, 'dBmW/Hz')
    assert str(q)                == "Quantity(1.000e+01 dBmW*Hz-1)"
    q.to('W/Hz')
    assert str(q)                == "Quantity(1.000e-02 W*Hz-1)"
    q = q*Quantity(100, 'Hz')
    assert str(q)                == "Quantity(1.000e+00 W)"
    q.to('dBm')
    assert str(q)                == "Quantity(3.000e+01 dBm)"
    
def test_logarithmic_arithmetics():
    
    a = Quantity(1, 'dB')
    b = Quantity(2, 'dB')
    assert str(a+b)     == "Quantity(4.539e+00 dB)"
    
    a = Quantity(87, 'dBA')
    b = Quantity(83, 'dBA')
    assert str(a-b)     == "Quantity(8.480e+01 dBA)"
    
    a = Quantity(20, 'dBm')
    b = Quantity(23, 'dBm')
    assert str(a+b)     == "Quantity(2.476e+01 dBm)"
    
def test_rebase():
    
    # simple prefix rebasing 
    assert str(Quantity(1, 'cm*m*dm').rebase())  == "Quantity(1.000e+03 cm3)"
    
    # rebasing prefixes with exponents
    assert str(Quantity(1, 'C3*cm*m2').rebase()) == "Quantity(1.000e+04 C3*cm3)"
    assert str(Quantity(1, 'cm*m3:2').rebase())  == "Quantity(1.000e+03 cm5:2)"
    
    # rebasing different units
    assert str(Quantity(1, 'erg*J').rebase())    == "Quantity(1.000e+07 erg2)"



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import sys
sys.path.insert(0,"../../src")              # import recent version of scinumtools
import scinumtools

import toml
data = toml.load("../../pyproject.toml")

project = data['project']['name']
copyright = '2023, '+data['project']['authors'][0]['name']
author = data['project']['authors'][0]['name']
release = data['project']['version']
version = data['project']['version']

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autosectionlabel',
    "sphinx.ext.autodoc", 
    'sphinx_rtd_theme',
    'sphinx_simplepdf',
]
autosectionlabel_prefix_document = True  # Make sure the target is unique
templates_path = ['_templates']
exclude_patterns = []



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']
html_logo = "_static/snt/snt.svg"
html_theme_options = {
    'logo_only': False,
    'display_version': True,
}
html_css_files = [
    'css/sphinxdoc.css'
]

# DIP Syntax Highlighter
from sphinx.highlighting import lexers
from scinumtools.dip.pygments import SyntaxLexer, SchemaLexer, StyleLexer, pygments_monkeypatch_style
pygments_monkeypatch_style("StyleLexer", StyleLexer)
pygments_style = 'StyleLexer'
lexers['DIP'] = SyntaxLexer(startinline=True, style=StyleLexer)
lexers['DIPSchema'] = SchemaLexer(startinline=True, style=StyleLexer)



import sys, os
import pytest
import numpy as np
sys.path.insert(0, 'src')

from scinumtools.dip import DIP
from scinumtools.dip.datatypes import FloatType

def test_node_list():
    with DIP() as p:
        p.add_file("examples/definitions.dip")
        env = p.parse()
        
    assert env.nodes.keys()                       == ['box', 'modules', 'runtime', 'simulation']
    assert env.nodes['runtime'].keys()            == ['t_max', 'timestep']
    assert env.nodes['runtime']['timestep'].value == FloatType(1.0000000000000001e-11, 's')
    assert env.nodes['runtime.t_max'].value       == FloatType(1e-08, 's')


a = """
from PyQt6.QtWidgets import (
    QApplication,
    QCheckBox,
    QComboBox,
    QDateEdit,
    QDateTimeEdit,
    QDial,
    QDoubleSpinBox,
    QFontComboBox,
    QLabel,
    QLCDNumber,
    QLineEdit,
    QMainWindow,
    QProgressBar,
    QPushButton,
    QRadioButton,
    QSlider,
    QSpinBox,
    QTimeEdit,
    QVBoxLayout,
    QWidget,
)

# Only needed for access to command line arguments
import sys

# Subclass QMainWindow to customize your application's main window
class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()

        self.setWindowTitle("Widgets App")

        layout = QVBoxLayout()
        widgets = [
            QCheckBox,
            QComboBox,
            QDateEdit,
            QDateTimeEdit,
            QDial,
            QDoubleSpinBox,
            QFontComboBox,
            QLCDNumber,
            QLabel,
            QLineEdit,
            QProgressBar,
            QPushButton,
            QRadioButton,
            QSlider,
            QSpinBox,
            QTimeEdit,
        ]

        for w in widgets:
            layout.addWidget(w())

        widget = QWidget()
        widget.setLayout(layout)

        # Set the central widget of the Window. Widget will expand
        # to take up all the space in the window by default.
        self.setCentralWidget(widget)


app = QApplication(sys.argv)
window = MainWindow()
window.show()

app.exec()
"""



import numpy as np
from dataclasses import dataclass, field #, fields
from typing import Union

from .fraction import Fraction
from .settings import DIMENSION_LIST
    
@dataclass
class Dimensions:
    
    m:   Fraction = field(default_factory=Fraction)
    g:   Fraction = field(default_factory=Fraction)
    s:   Fraction = field(default_factory=Fraction)
    K:   Fraction = field(default_factory=Fraction)
    C:   Fraction = field(default_factory=Fraction)
    cd:  Fraction = field(default_factory=Fraction)
    mol: Fraction = field(default_factory=Fraction)
    rad: Fraction = field(default_factory=Fraction)
    nodim: bool = field(default=True)
    
    @staticmethod
    def from_list(value: list):
        units = {}
        for n, name in enumerate(DIMENSION_LIST):
            if isinstance(value[n], tuple):
                units[name] = Fraction.from_tuple(value[n])
            else:
                units[name] = Fraction(value[n])
        return Dimensions(**units)
    
    def __post_init__(self):
        # are all dimensions zero?
        for name in DIMENSION_LIST:
            if getattr(self, name).num!=0:
                self.nodim = False

    def __str__(self):
        dimensions = []
        for name in DIMENSION_LIST:
            value = getattr(self, name)
            if value.num not in [0, -0]:
                dimensions.append(f"{name}={str(value)}")
        dimensions = " ".join(dimensions)
        return f"Dimensions({dimensions})"

    def __repr__(self):
        dimensions = []
        for name in DIMENSION_LIST:
            value = getattr(self, name)
            if value.num not in [0, -0]:
                dimensions.append(f"{name}={str(value)}")
        dimensions = " ".join(dimensions)
        return f"Dimensions({dimensions})"
                            
    def __add__(self, other):
        dimensions = {}
        for name in DIMENSION_LIST:
            if isinstance(other, Dimensions):
                dimensions[name] = getattr(self, name) + getattr(other, name)
            else:
                dimensions[name] = getattr(self, name) + other
        return Dimensions(**dimensions)
            
    def __sub__(self, other):
        dimensions = {}
        for name in DIMENSION_LIST:
            if isinstance(other, Dimensions):
                dimensions[name] = getattr(self, name) - getattr(other, name)
            else:
                dimensions[name] = getattr(self, name) - other
        return Dimensions(**dimensions)

    def __mul__(self, other):
        dimensions = {}
        for name in DIMENSION_LIST:
            dimensions[name] = getattr(self, name) * other
        return Dimensions(**dimensions)

    def __truediv__(self, other):
        dimensions = {}
        for name in DIMENSION_LIST:
            dimensions[name] = getattr(self, name) / other
        return Dimensions(**dimensions)
    
    def __eq__(self, other):
        for name in DIMENSION_LIST:
            if not getattr(self, name) == getattr(other, name):
                return False
        return True

    def __neg__(self):
        """ Inverse dimensions
        """
        dimensions = {}
        for name in DIMENSION_LIST:
            dimensions[name] = getattr(self, name) * -1
        return Dimensions(**dimensions)

    def value(self, dtype=list):
        if dtype==list:
            dimensions = []
            for name in DIMENSION_LIST:
                dimensions.append( getattr(self, name).value() )
        elif dtype==dict:
            dimensions = {}
            for name in DIMENSION_LIST:
                fraction = getattr(self, name)
                if fraction.num not in [0, -0]:
                    dimensions[name] = fraction.value()
        elif dtype==tuple:
            dimensions = []
            for name in DIMENSION_LIST:
                fraction = getattr(self, name)
                if fraction.num not in [0, -0]:
                    dimensions.append(name)
            dimensions = tuple(dimensions)
        return dimensions
        


from PIL import Image, ExifTags
import json
from enum import Enum

class Metadata(Enum):
    DIR_SETUP    = 'DirSetup'
    DIR_DATA     = 'DirData'
    GIT_COMMIT   = 'GitCommit'
    GIT_BRANCH   = 'GitBranch'
    DATETIME     = 'DateTime'
    SETTINGS     = 'Settings'

class ImageMetadata:
    
    modified:bool = False
    file_name:str
    metadata:dict
    
    def __enter__(self):
        return self
        
    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.modified:
            self.save()
    
    def __init__(self, file_name:str):
        self.file_name = file_name
        with Image.open(file_name) as img:
            self.metadata = {}
            if 'exif' in img.info:
                exif = img.getexif()
                self.metadata = json.loads(exif[0x9286]) # UserComment tag

    def __setitem__(self, key:str, value:str):
        self.set(key,value)
        
    def __getitem__(self, key:str):
        return self.get(key)

    def set(self, key:Metadata, value:str):
        self.metadata[key.value] = value
        self.modified = True
        
    def get(self, key:Metadata):
        return self.metadata[key.value]
        
    def save(self):
        img = Image.open(self.file_name)
        exif = img.getexif()
        exif[0x9286] = json.dumps(self.metadata) # UserComment tag
        img.save(self.file_name, img.format, exif=exif)
        
    def print(self):
        text = []
        for key,value in self.metadata.items():
            text.append(f"{key}: {self.metadata[key]}")
        print("\n".join(text))


import numpy as np
import pytest
from math import isclose
import os
import sys
sys.path.insert(0, 'src')

from scinumtools.units import *
from scinumtools.units.systems import SI
from scinumtools.units.unit_environment import *
    
def test_unit_list():
    
    assert str(Unit())
    assert str(Constant())

    #print(Constant())
    #print(Unit())
    #exit(1)
    
def test_unique_unique():
    
    assert check_unique_symbols()

def test_custom_units():
    
    # register custom unit
    units = {
        'x': {'magnitude':3, 'dimensions':[3,2,-1,0,0,1,0,0]},   # defined as a dictionary
        'y': Quantity(2, 'cm/g2')                                # defined as a quantity
    }
    with UnitEnvironment(units):
    
        q = Quantity(1, 'x')
        assert str(q) == "Quantity(1.000e+00 x)"
        assert q.baseunits.magnitude          == 3.0
        assert q.baseunits.dimensions.value() == [3,2,-1,0,0,1,0,0]
        
        q = Quantity(1, 'y')
        assert str(q) == "Quantity(1.000e+00 y)"
        assert q.baseunits.magnitude          == 0.02
        assert q.baseunits.dimensions.value() == [1,-2,0,0,0,0,0,0]

        with pytest.raises(Exception) as excinfo:
            Quantity(1, 'kx')
        assert excinfo.value.args[0]=="Unit cannot have any prefixes:"
        assert excinfo.value.args[1]=="x"

        # try to register new unit with already existing symbol
        with pytest.raises(Exception) as excinfo:
            with UnitEnvironment(units):
                pass
        assert excinfo.value.args[0]=="Unit with this symbol already exists:"
        assert excinfo.value.args[1]=="x"
    
    # make sure that units does not exist outside of the environment
    assert 'x' not in UNIT_STANDARD
    
    # register new unit with a custom converter 
    class CustomUnitType(UnitType):
        def _istype(self):
            return False
    units = {'x': {'magnitude':3, 'dimensions':[3,2,-1,0,0,1,0,0], 'definition':CustomUnitType}}
    env = UnitEnvironment(units)
    assert 'x' in UNIT_STANDARD
    assert CustomUnitType in UNIT_TYPES
    env.close()
    
    # make sure that units and types does not exist aftr environment is closed
    assert 'x' not in UNIT_STANDARD
    assert CustomUnitType not in UNIT_TYPES
    

def test_prefixes():
    
    # unit can have any prefix
    assert str(Quantity(1, 'km'))    == "Quantity(1.000e+00 km)"
    
    # unit cannot have prefixes
    with pytest.raises(Exception) as excinfo:
        Quantity(1, 'kCel')
    assert excinfo.value.args[0]=="Unit cannot have any prefixes:"
    assert excinfo.value.args[1]=="Cel"
    
    # unit can have only selected prefixes
    with pytest.raises(Exception) as excinfo:
        Quantity(1, 'mpc')
    assert excinfo.value.args[0]=="Unit can have only following prefixes:"
    assert excinfo.value.args[1]==['k', 'M', 'G', 'T']
    

def test_definitions():
    
    for symbol, unit in UNIT_STANDARD.items():
        if not isinstance(unit.definition,str):
            continue
        q = Quantity(1, unit.definition)
        base = q.baseunits
        magnitude = q.magnitude.value*base.magnitude
        if base.dimensions.value(dtype=list) != unit.dimensions or not isclose(magnitude,  unit.magnitude, rel_tol=MAGNITUDE_PRECISION):
            print(q, symbol, unit.definition, unit.magnitude)
        assert isclose(magnitude,  unit.magnitude, rel_tol=MAGNITUDE_PRECISION)
        assert base.dimensions.value(dtype=list) == unit.dimensions
        
    for symbol, unit in UNIT_PREFIXES.items():
        if not isinstance(unit.definition,str):
            continue
        q = Quantity(1, unit.definition)
        base = q.baseunits
        assert isclose(q.magnitude.value,  unit.magnitude, rel_tol=MAGNITUDE_PRECISION)
        assert base.dimensions.value(dtype=list) == unit.dimensions
        
def test_unit_systems():

    # get the quantity symbols
    assert SI['Pressure']           == '#SPRE'
    assert CGS['Energy']            == '#CENE'
    assert AU['ElectricCharge']     == '#AECH'

    # test if it works with Quantity
    assert str(Quantity(23, SI['Pressure']))          == 'Quantity(2.300e+01 #SPRE)'
    assert str(Quantity(23, 'Pa').to(SI['Pressure'])) == 'Quantity(2.300e+01 #SPRE)'
    assert str(Quantity(23, SI['Pressure']).to('Pa')) == 'Quantity(2.300e+01 Pa)'
    assert str(Quantity(1, '[a_0]').to(AU['Length'])) == 'Quantity(1.000e+00 #ALEN)'
    assert str(Quantity(1, '#ALEN/s').to('m/s'))      == 'Quantity(5.292e-11 m*s-1)'

    # test if it works with Unit
    assert str(Unit(AU['Length']))                    == "Quantity(1.000e+00 #ALEN)"
    assert str(Unit('[a_0]')/Unit(AU['Length']))      == 'Quantity(1.000e+00)'
    assert str(Unit(AU['Length'])*Unit('s'))          == 'Quantity(1.000e+00 #ALEN*s)'

    # test if arguments create a quantity
    assert str(AU.Pressure())            == "Quantity(1.000e+00 #APRE)"  # equivalent to Unit
    assert str(SI.Pressure(23))          == "Quantity(2.300e+01 #SPRE)"  # equivalent to Quantity
    assert str(CGS.Pressure(23))         == "Quantity(2.300e+01 #CPRE)"



from typing import List
import copy

class Node:
    code: str 
    source: tuple = None            # source 
    isource: tuple = False          # import source
    keyword: str = None
    dtype = str                     # datatype
    dtype_prop = list               # datatype properties (e.g.: unsigned, precision)
    indent: int = 0
    name: str = None                # node name
    value_raw: str = None           # Raw value
    value_ref: str = None           # Reference
    value_fn: str = None            # Function
    value_expr: str = None          # Expression
    value_slice: List[tuple] = None # Slice
    units_raw: str = None           # Ras units
    defined: bool = False
    dimension: List[tuple] = None

    def __init__(self, code, **kwargs):
        self.code = code
        self.dtype_prop = []
        for key,val in kwargs.items():
            setattr(self, key, val)

    def copy(self):
        return copy.deepcopy(self)


import numpy as np
import toml

from .export import ExportConfig
from ..settings import Format
from ..environment import Environment

class ExportConfigTOML(ExportConfig):

    def __init__(self, env: Environment, **kwargs):
        if 'dtype' not in kwargs:
            kwargs['dtype'] = Format.TUPLE
        super().__init__(env, **kwargs)

    def parse(self, units:bool = True, **kwargs):
        data = self.data
        for key in self.data.keys():
            if isinstance(data[key], tuple):
                if units:
                    data[key] = {'value': data[key][0], 'unit':data[key][1]}
                else:
                    data[key] = data[key][0]
        self.text = toml.dumps(data, **kwargs).strip()
        return self.text


from dataclasses import dataclass, field
from typing import List, Dict, Union

@dataclass
class FunctionList:
    functions: Dict = field(default_factory = dict)
    
    def __getitem__(self, key: str):
        return self.functions[key]
            
    def append(self, name, function):
        self.functions[name] = function


from .settings import *
from ..settings import ROOT_SOURCE

class SourceItem:
    target: str
    link_source: str
    name: str
    path: str
    parent: tuple
    code: str
            
    def __init__(self, source):
        self.target = SourceItemTarget(source.name, None)
        if source.parent:
            self.link_source = SourceItemTarget(source.parent[0], source.parent[1])
        self.name   = source.name
        self.path   = source.path
        self.parent = source.parent
        self.code   = source.code




from dataclasses import dataclass, field
from typing import List, Dict, Union
import numpy as np
import re

from ..settings import Sign, Keyword
from ..datatypes import BooleanType

@dataclass
class Case:
    path: str        # case path up to last @ sign
    value: bool      # final value of the case
    code: str        # code line with the case
    expr: str        # case expression
    branch_id: str   # branch ID
    branch_part: str # branch part
    case_id: str     # case ID
    case_type: str   # one of the types: case/else/end

@dataclass
class Branch:
    cases: list   = field(default_factory = list)  # list of case IDs
    types: list   = field(default_factory = list)  # list of case types
    nodes: dict   = field(default_factory = dict)  # number of node definitions
    
@dataclass
class BranchingList:
    # branch structure
    state: List[str]           = field(default_factory = list)  # list of openned branches
    branches: Dict[str,Branch] = field(default_factory = dict)  # all branches
    cases: Dict[str,Case]      = field(default_factory = dict)  # all cases

    # counters
    num_cases: int = 0
    num_branches: int = 0

    def _get_branch_id(self):
        """ Get ID of a current branch
        """
        if self.state:
            return self.state[-1]
        else:
            return None

    def _get_case_id(self):
        """ Get ID of a current case
        """
        branch_id = self._get_branch_id()
        return self.branches[branch_id].cases[-1]
        
    def _open_branch(self, case_id):
        """ Start a new branch
        """
        self.num_branches += 1        
        branch_id = f"{Sign.CONDITION}{self.num_branches}"
        self.state.append(branch_id)
        self.branches[branch_id] = Branch([case_id], [Keyword.CASE])
        return 0  # branch_part
    
    def _switch_case(self, case_id, case_type):
        """ Go to a new case withing a branch
        """
        branch_id = self._get_branch_id()
        self.branches[branch_id].cases.append(case_id)
        self.branches[branch_id].types.append(case_type)
        return len(self.branches[branch_id].cases)-1  # branch_part

    def _close_branch(self):
        """ Close current branch
        """
        #if len(self.state)>1:
        #    print(self.state[-1])
        # remove current branch from the state list
        self.state.pop()
       
    def register_case(self):
        """ Add a new case
        """
        self.num_cases += 1
        return self.num_cases
    
    def false_case(self):
        """ Checks if case value is false
        """
        if not self.state:
            return False
        # count number of true cases
        branch = self._get_branch_id()
        num_true = sum([self.cases[c].value==True for c in self.branches[branch].cases])
        # only first `true` case is valid
        case = self._get_case_id()
        return num_true!=1 or self.cases[case].value == False
        
    def solve_case(self, node):
        """ Manage condition nodes

        :param node: Condition node
        """
        if m := re.match(f"(.*{Sign.CONDITION})([0-9]+)$", node.name):
            path_new = m.group(1)
            path_old = ''
            if self.state:
                id_old = self._get_case_id()
                path_old = self.cases[id_old].path
            if node.case_type==Keyword.CASE:
                pass
            elif node.case_type==Keyword.ELSE and self.cases:
                pass
            elif node.case_type==Keyword.END and self.cases and path_old==path_new:
                self._close_branch()
                return
            else:
                raise Exception(f"Invalid condition:", node.code)
            case_id = fr"{Sign.CONDITION}{m.group(2)}"
            if path_new==path_old:  # same branch
                branch_part = self._switch_case(case_id, node.case_type)
            elif path_new<path_old: # lower branch
                # close openned branches unitil the same branch is reached
                while path_new!=path_old:
                    self._close_branch()
                    if self.state:
                        id_old = self._get_case_id()
                        path_old = self.cases[id_old].path
                    else:
                        path_old = ''
                branch_part = self._switch_case(case_id, node.case_type)
            else:                   # new branch
                branch_part = self._open_branch(case_id)
            branch_id = self._get_branch_id()
            self.cases[case_id] = Case(
                path        = path_new,          # path of a new case
                code        = node.code,         # code line
                expr        = node.value_expr,   # case logical expression
                value       = node.value,        # boolean value true/false
                branch_id   = branch_id,         # branch ID
                branch_part = branch_part,       # part on the branch
                case_id     = case_id,           # case ID
                case_type   = node.case_type,    # case type CASE/ELSE/END
            )
        else:
            raise Exception(f"Invalid condition:", node.code)

    def prepare_node(self, node):
        """ Manage parameter nodes in a condition

        :param node: Parameter node
        """
        if not self.state: # outside of any condition
            return
        case = self._get_case_id()
        if not node.name.startswith(self.cases[case].path): # ending case at lower indent
            self._close_branch()
        if self.state:
            node.branch_id = self._get_branch_id()
            node.case_id   = self._get_case_id()
            # Register node to the branch
            branch_id = self._get_branch_id()
            node_name = node.clean_name()
            if node_name in self.branches[branch_id].nodes:
                self.branches[branch_id].nodes[node_name] += 1
            else:
                self.branches[branch_id].nodes[node_name] = 1 



import numpy as np
import pandas as pd
from textwrap import dedent
import sys
sys.path.insert(0, 'src')

import scinumtools as snt

def test_parameter_list():
    def test(params):
        assert params[0].a == 1
        assert params[1]['a'] == 4
        for param in params:
            assert param['a'] in [1,4]
        for index,settings in params.items():
            assert param['b'] in [2,5]        
    # append values
    with snt.ParameterTable(['a','b','c']) as params:
        params.append([1, 2, 3])
        params.append([4, 5, 6])
        test(params)
    # direct value insertion
    params = snt.ParameterTable(['a','b','c'],[
        [1, 2, 3],
        [4, 5, 6]
    ])
    test(params)

def test_parameter_dict():
    def test(params):
        assert params['d'].a == 1
        assert params['e']['a'] == 4
        for param in params:
            assert param['a'] in [1,4]
        for index,settings in params.items():
            assert param['b'] in [2,5]
    # append values
    with snt.ParameterTable(['a','b','c'], keys=True) as params:
        params['d'] = [1, 2, 3]
        params.append( 'e', [4, 5, 6] )
        test(params)
    # direct value insertion
    params = snt.ParameterTable(['a','b','c'],{
        'd': [1, 2, 3],
        'e': [4, 5, 6]
    }, keys=True)
    test(params)
    
def test_shape():
    
    with snt.ParameterTable(['a','b','c']) as params:
        params.append([1, 2, 3])
        params.append([4, 5, 6])
        assert params.shape() == (2, 3)
        
def test_rows_attributes():
    
    with snt.ParameterTable(['a','b','c'], keys=True) as params:
        params['d'] = [1, 2, 3]
        params.append( 'e', [4, 5, 6] )
        
        assert str(params.d)   == "ParameterSettings(a=1 b=2 c=3)"
        assert params.d.a      == 1
        
def test_delete_item():
    
    with snt.ParameterTable(['a','b','c']) as params:
        params.append([1, 2, 3])
        params.append([4, 5, 6])
        
        assert len(params) == 2
        del params[0]
        assert len(params) == 1
        assert params[0].a == 4
        assert params[0].b == 5
        assert params[0].c == 6
        
    with snt.ParameterTable(['a','b','c'], keys=True) as params:
        params['d'] = [1, 2, 3]
        params.append( 'e', [4, 5, 6] )
        
        del params['d']
        assert params._keys == ['e']
        assert 'd' not in params._data
        assert params['e'].a == 4
        assert params['e'].b == 5
        assert params['e'].c == 6

def test_data():
    
    with snt.ParameterTable(['a','b','c']) as params:
        params.append([1, 2, 3])
        params.append([4, 5, 6])
        assert params.data() == [
            {'a': 1, 'b': 2, 'c': 3},
            {'a': 4, 'b': 5, 'c': 6},
        ]
        
    with snt.ParameterTable(['a','b','c'], keys=True) as params:
        params['d'] = [1, 2, 3]
        params['e'] = [4, 5, 6]
        assert params.data() == {
            'd': {'a': 1, 'b': 2, 'c': 3},
            'e': {'a': 4, 'b': 5, 'c': 6},
        }
        
def test_to_dataframe():
    
    with snt.ParameterTable(['a','b','c']) as params:
        params.append([1, 2, 3])
        params.append([4, 5, 6])
        assert params.to_dataframe().equals(pd.DataFrame({
            'a': [1, 4],
            'b': [2, 5],
            'c': [3, 6],
        }))
        
    with snt.ParameterTable(['a','b','c'], keys=True) as params:
        params['d'] = [1, 2, 3]
        params['e'] = [4, 5, 6]
        assert params.to_dataframe().equals(pd.DataFrame({
            '#': ['d','e'],
            'a': [1, 4],
            'b': [2, 5],
            'c': [3, 6],
        }))
        
def test_to_text():
    
    with snt.ParameterTable(['a','b','c']) as params:
        params.append([1, 2, 3])
        params.append([4, 5, 6])
        assert params.to_text() == """
   a  b  c
0  1  2  3
1  4  5  6
""".strip('\n')
        
    with snt.ParameterTable(['a','b','c'], keys=True, keyname='x') as params:
        params['d'] = [1, 2, 3]
        params['e'] = [4, 5, 6]
        assert params.to_text() == """
   x  a  b  c
0  d  1  2  3
1  e  4  5  6
""".strip('\n')

def test_print():
    pt = snt.ParameterTable(['a','b','c'],{
        'd': [1, 2, 3],
        'e': [4, 5, 6]
    }, keys=True)
    assert str(pt) == 'ParameterSettings(d, e)'


from enum import Flag, auto

from ..settings import ROOT_SOURCE, Sign

class DocsType(Flag):  # node type in documentation
    UNKNOWN      = auto()
    DECLARATION  = auto()
    DEFINITION   = auto()
    MODIFICATION = auto()
    INJECTION    = auto()
    
def NormalizeTargetName(name: str):
    name = name.replace('.','_')
    name = name.replace(Sign.CONDITION,'CASE')
    return name
    
def ParameterItemTarget(name: str):
    name = NormalizeTargetName(name)
    return f"PARAM_{name}"
    
def NodeItemTarget(name:str, source:str, lineno:int):
    name = NormalizeTargetName(name)
    return f"NODE_{name}_{source}_{lineno}"
    
def InjectionItemTarget(name:str, source:str, lineno:int):
    name = NormalizeTargetName(name)
    return f"INJECT_{name}_{source}_{lineno}"

def ImportItemTarget(source:str, lineno:int):
    return f"IMPORT_{source}_{lineno}"

def UnitItemTarget(name: str):
    name = NormalizeTargetName(name)
    return f"UNIT_{name}"
    
def SourceItemTarget(source:str, lineno:int=None):
    if ROOT_SOURCE in source or lineno is None:
        return f"SOURCE_{source}"
    else:
        return f"SOURCE_{source}_{lineno}"
        


from .settings import *
from dataclasses import dataclass, field

@dataclass
class ImportItemNode:
    name: str
    source: tuple
    link_node: str = None
    link_source: str = None

    def __post_init__(self):
        self.link_node = NodeItemTarget(self.name, self.source[0], self.source[1])
        self.link_source = SourceItemTarget(self.source[0], self.source[1])
        
class ImportItem:
    target: str
    link_source: str
    name: str
    reference: str
    source: tuple
    idata: list
    
    def __init__(self, node, nodes):
        self.target = ImportItemTarget(node.source[0], node.source[1])
        self.link_source = SourceItemTarget(node.source[0], node.source[1])
        self.name = node.clean_name().split('.{')[0]
        self.source = node.source
        self.reference = node.value_ref
        self.idata = []
        for inode in nodes:
            if inode.isource==node.source:
                self.idata.append(ImportItemNode(inode.name,inode.source))
    


import numpy as np
from typing import List, Dict, Union
from dataclasses import dataclass, field
import copy

from .settings import *
from .item_parameter import ParameterItem, ParType
from .item_injection import InjectionItem
from .item_import import ImportItem
from .item_unit import UnitItem
from .item_source import SourceItem
from ..environment import Environment
from ..nodes import ModNode, ImportNode, StringNode
from ..settings import Sign, Keyword, EnvType

@dataclass
class Documentation:
    
    env: Environment
    types: dict      = field(default_factory=dict)
    parameters: dict = field(default_factory=dict)
    imports: list    = field(default_factory=list)
    injections: list = field(default_factory=list)
    units: list      = field(default_factory=list)
    sources: list    = field(default_factory=list)
    
    def __post_init__(self):
        if self.env.envtype != EnvType.DOCS:
            raise Exception("Given environment is not a documentation environment")
        # group nodes according to their names
        nodes = self.env.nodes.query("*")
        
        # populate sources
        self.sources = []
        for name,source in self.env.sources.items():
            self.sources.append(SourceItem(source))
        
        # populate units
        self.units = []
        for name,unit in self.env.units.items():
            self.units.append(UnitItem(name, unit))
        
        # populate nodes, parameters, injections and imports
        self.parameters = {}
        self.injections = []
        self.imports = []
        for node in nodes:
            if node.keyword==ImportNode.keyword:
                self.imports.append(ImportItem(node, nodes))
            else:
                name = node.clean_name()
                if name in self.parameters:
                    self.parameters[name].add(name, node)
                else:
                    self.parameters[name] = ParameterItem(name, node)
                if node.value_ref:
                    self.injections.append(InjectionItem(node,self.env))
                    
        # populate node types
        self.types = [t.value for t in ParType]
    
    def copy(self):
        """ Copy a new object from self
        """
        return copy.deepcopy(self)


import numpy as np
import os
import sys
sys.path.insert(0, 'src')
import pytest
import glob

import scinumtools as snt

@pytest.fixture()
def file_cache():
    dir_cache = "tmp"
    name_cache = "cached_data"
    if os.path.isdir(dir_cache):
        for item in os.listdir(dir_cache):
            if item.startswith(name_cache):
                os.remove(f"{dir_cache}/{item}")
    else:
        os.mkdir(dir_cache)
    return f"{dir_cache}/{name_cache}.npy"

def cache_exists(file_cache, *args, **kwargs):
    file_data = snt.cached_function.hash_file_name(file_cache, args, kwargs)
    return True if glob.glob(file_data) else False

def test_arguments(file_cache):
    
    @snt.CachedFunction(file_cache)
    def read_data(a, b):
        return dict(a=a, b=b)
    
    data = read_data('foo','bar')    
    assert data == dict(a='foo', b='bar')

    # same cache when arguments are the same
    data = read_data('foo','bar')
    assert cache_exists(file_cache,'foo','bar')
    assert data == dict(a='foo', b='bar')
    
    # new cache when arguments are different
    data = read_data('foo2','bar2')
    assert cache_exists(file_cache,'foo2','bar2')
    assert data == dict(a='foo2', b='bar2')

def test_caching(file_cache):

    @snt.CachedFunction(file_cache)
    def read_data():
        return {
            'array': np.linspace(1,100,100),
            'scalar': 34,
            'string': 'foo',
        }
    
    data = read_data()
    np.testing.assert_equal(data, dict(
        array=np.linspace(1,100,100),
        scalar=34,
        string='foo',
    ))



import numpy as np
import pytest
from math import isclose
import os
import pandas as pd
import sys
import time
import cProfile
import pstats
sys.path.insert(0, '../../../src')

from scinumtools.units import *
from scinumtools import RowCollector

if __name__ == '__main__':
    
    file_csv = "material_layers.csv"
    
    with RowCollector(['layer', 'dx', 'T_i', 'T_e', 'rho', 'v']) as rc:
        for i in range(100):
            rc.append(['Li',   20, 20e3, 20e3, 0.534, 0])
            rc.append(['Al',   20, 20e3, 20e3, 2.7,   0])
        for i in range(40):
            rc.append(['Au',   50, 3e-3, 3e-3, 19.32, 0])
        df = rc.to_dataframe()
    df.to_csv(file_csv)

    with cProfile.Profile() as pr:

        height  = quant(1, 'mm').value('cm')
        mass_Au  = quant(10.811, 'Da').value('g')
        mass_Al  = quant(1.0078, 'Da').value('g')
        mass_Li  = quant(2.01410177811, 'Da').value('g')

        csv = pd.read_csv(file_csv)
        rmin = quant(0, 'mm')
        numAu, numAl, numLi = 0, 0, 0
        for index, row in csv.iterrows():
            rmax = rmin + quant(row.dx, 'cm')
            volume = height * np.pi * (rmax**2 - rmin**2)
            mass = volume * quant(row.rho, 'g/cm3')
            layer = row.layer.strip()
            if layer.endswith('Al'):
                numAl = (mass/mass_Al).rebase().value()
            elif layer.endswith('Au'):
                numAu = (mass/mass_Au).rebase().value()
            elif layer.endswith('Li'):
                numLi += (mass/mass_Li).rebase().value()
            rmin = rmax

    stats = pstats.Stats(pr)
    stats.sort_stats(pstats.SortKey.TIME)
    stats.print_stats(20)



import numpy as np
import pytest
from math import isclose
import os
import sys
sys.path.insert(0, 'src')

from scinumtools.units import Quantity, Unit 
from scinumtools.materials import *

def test_preprocessing():
    
    compound = Substance()
    with SubstanceSolver(compound.atom) as ms:
        
        # implicit multiplication and addition
        assert ms.preprocess('C2B4')  == "C * 2 + B * 4"
        assert ms.preprocess('C2 B4') == "C * 2 + B * 4"
        # isotopes and ionizations
        assert ms.preprocess('C{13+2}B{11}H{-}2') == "C{13+2} + B{11} + H{-} * 2"
        # parentheses
        assert ms.preprocess('(CB2)2')                         == "(C + B * 2) * 2"
        assert ms.preprocess('((CB2)2Al)3')                    == "((C + B * 2) * 2 + Al) * 3"
        assert ms.preprocess('C{13+2}(B{11}Li2)4 H{-}2 O{+3}') == "C{13+2} + (B{11} + Li * 2) * 4 + H{-} * 2 + O{+3}"
    
        # preprocessed expressions are constant
        assert ms.preprocess("(C + B * 2) * 2")             == "(C + B * 2) * 2"
        assert ms.preprocess("C{13+2} + (B{11} + Li * 2)4") == "C{13+2} + (B{11} + Li * 2) * 4"

def test_solver():
    
    compound = Substance()
    with SubstanceSolver(compound.atom) as ms:
        
        # single atom parsing
        assert str(ms.solve('C'))       == "Substance(mass=12.011 Z=6 N=6.011 e=6)"
        assert str(ms.solve('C{12}'))   == "Substance(mass=12.000 Z=6 N=6.000 e=6)"
        assert str(ms.solve('C{-2}'))   == "Substance(mass=12.010 Z=6 N=6.011 e=4)"
        assert str(ms.solve('C{+}'))    == "Substance(mass=12.011 Z=6 N=6.011 e=7)"
        assert str(ms.solve('C{12-}'))  == "Substance(mass=11.999 Z=6 N=6.000 e=5)"
        assert str(ms.solve('C{13}'))   == "Substance(mass=13.003 Z=6 N=7.000 e=6)"
        assert str(ms.solve('C{13-2}')) == "Substance(mass=13.002 Z=6 N=7.000 e=4)"

        # nucleons
        assert str(ms.solve('[e]'))      == "Substance(mass=0.001 Z=0 N=0.000 e=1)"
        assert str(ms.solve('[n]'))      == "Substance(mass=1.009 Z=0 N=1.000 e=0)"
        assert str(ms.solve('[n]2'))     == "Substance(mass=2.017 Z=0 N=2.000 e=0)"
        assert str(ms.solve('[p]'))      == "Substance(mass=1.007 Z=1 N=0.000 e=0)"
        assert str(ms.solve('[p]2'))     == "Substance(mass=2.015 Z=2 N=0.000 e=0)"
        assert str(ms.solve('[p]B{11}')) == "Substance(mass=12.017 Z=6 N=6.000 e=5)"
        
        # hydrogen isotopes
        assert str(ms.solve('[p]'))     == str(ms.solve('H{1-1}'))
        assert str(ms.solve('H{1+}'))   == str(ms.solve('H{1+1}'))
        assert str(ms.solve('D'))       == str(ms.solve('H{2}'))
        assert str(ms.solve('D{2-2}'))  == str(ms.solve('H{2-2}'))
        assert str(ms.solve('T'))       == str(ms.solve('H{3}'))
    
        # addition       
        assert str(ms.solve('H{1-1}'))           == "Substance(mass=1.007 Z=1 N=0.000 e=0)"
        assert str(ms.solve('B{11}'))            == "Substance(mass=11.009 Z=5 N=6.000 e=5)"
        assert str(ms.solve('H{1-1} + B{11}'))   == "Substance(mass=12.017 Z=6 N=6.000 e=5)"
    
        # parentheses
        assert str(ms.solve('(H2O)'))   == "Substance(mass=18.015 Z=10 N=8.005 e=10)"
        assert str(ms.solve('H(CN)'))   == "Substance(mass=27.025 Z=14 N=13.014 e=14)"
        assert str(ms.solve('Ca(OH)2')) == "Substance(mass=74.093 Z=38 N=36.125 e=38)"

        # multiplication
        assert str(ms.solve('H{1-1}2'))             == "Substance(mass=2.015 Z=2 N=0.000 e=0)"
        assert str(ms.solve('(H{1-1} + B{11})2'))   == "Substance(mass=24.033 Z=12 N=12.000 e=10)"
        


from pygments.lexer import RegexLexer, include, bygroups
from pygments.token import Token
    
class SyntaxLexer(RegexLexer):
    name = 'DIP'
    tokens = {
        'root': [
            (r'([ ]*)(#[^\n]*|)(\n)', 
             bygroups(Token.DIP.Text, Token.DIP.Comment, Token.DIP.Text)),
            (r'([ ]*)(\$source|\$unit)([ ]+)({)',
             bygroups(Token.DIP.Text, Token.DIP.Keyword, Token.DIP.Text,
                      Token.DIP.Reference),'reference'),
            (r'([ ]*)(\$source|\$unit)([ ]+)([a-zA-Z0-9-_]+)([ ]+=)',
             bygroups(Token.DIP.Text, Token.DIP.Keyword, Token.DIP.Text,
                      Token.DIP.Name, Token.DIP.Text),'node_value'),
            (r'[ ]*\@case', Token.DIP.Keyword, 'node_value'),
            (r'[ ]*\@(else|end)', Token.DIP.Keyword),
            (r'[ ]*!(options|format|condition|constant|tags|description)',
             Token.DIP.Keyword, 'node_value'), # full keywords
            (r'[ ]*!(desc)',
             Token.DIP.Keyword, 'node_value'), # abbreviated forms
            (r'^([ ]*)(=)',
             bygroups(Token.DIP.Text, Token.DIP.Keyword), 'node_value'),
            (r'^([ ]*)([a-zA-Z0-9-_.]+)([ ]+=)',
             bygroups(Token.DIP.Text, Token.DIP.Name, Token.DIP.Text), 'node_value'),
            (r'^([ ]*)([a-zA-Z0-9-_.]+)',
             bygroups(Token.DIP.Text, Token.DIP.Name), 'node_type'),
            (r'([ ]*)({)',
             bygroups(Token.DIP.Text, Token.DIP.Reference), 'reference'),
        ],
        'node_type': [
            (r'([ ]*)({)',
             bygroups(Token.DIP.Text, Token.DIP.Reference), 'reference'),
            (r'([ ]*)(bool|[u]?int[0-9]*|float[0-9]*|str|table)(\[)',
             bygroups(Token.DIP.Text, Token.DIP.Type, Token.DIP.Type), 'type_dim'),
            (r'([ ]*)(bool|[u]?int[0-9]*|float[0-9]*|str|table)([ ]+=)',
             bygroups(Token.DIP.Text, Token.DIP.Type, Token.DIP.Text), 'node_value'),
            (r'([ ]*)(bool|[u]?int[0-9]*|float[0-9]*|str|table)', 
             bygroups(Token.DIP.Text, Token.DIP.Type), 'unit'), 
            (r'([ ]*)(bool|[u]?int[0-9]*|float[0-9]*|str|table)',
             bygroups(Token.DIP.Text, Token.DIP.Type), '#pop'),
            (r' ', Token.DIP.Text, '#pop'),
        ],
        'type_dim': [
            (f'[0-9.:]+',              Token.DIP.Dimension),
            (f',',                     Token.DIP.Type),
            (r'(\])([ ]+=)',
             bygroups(Token.DIP.Type, Token.DIP.Text), 'node_value'),
            (r'\]',                    Token.DIP.Type, 'unit'),
            (r' ', Token.DIP.Text, '#pop:2'),
        ],
        'node_value': [
            # booleans
            (r"([ ]*)(true|false)",
             bygroups(Token.DIP.Text, Token.DIP.Boolean)),
            # numbers
            (r"([ ]*)(-[0-9.]+|[0-9.]+)(e-[0-9]+|e[0-9]+|)([ ]+[^#\n ]+|)",
             bygroups(Token.DIP.Text, Token.DIP.Number, Token.DIP.Number, Token.DIP.Unit)),
            # strings
            (r"([ ]*)(')",
             bygroups(Token.DIP.Text, Token.DIP.String),   'str_single'),
            (r'([ ]*)(""")',
             bygroups(Token.DIP.Text, Token.DIP.String),   'str_triple'),            
            (r'([ ]*)(")',
             bygroups(Token.DIP.Text, Token.DIP.String),   'str_double'),
            # function
            (r"([ ]*\()([a-zA-Z0-9_-]+)(\))",
             bygroups(Token.DIP.Text, Token.DIP.Expression, Token.DIP.Text), 'unit'),
            # expressions
            (r"([ ]*\()(')",
             bygroups(Token.DIP.Text, Token.DIP.Expression),   'expr_single'),
            (r'([ ]*\()(""")',
             bygroups(Token.DIP.Text, Token.DIP.Expression),   'expr_triple'),            
            (r'([ ]*\()(")',
             bygroups(Token.DIP.Text, Token.DIP.Expression),   'expr_double'),
            # references
            (r'([ ]*)({)',
             bygroups(Token.DIP.Text, Token.DIP.Reference), 'reference'),
            # arrays
            (r"([ ]*)(\[)",
             bygroups(Token.DIP.Text, Token.DIP.Text), '#push'),
            (r"(,)", Token.DIP.Text),
            (r"(\])([ ]+[^#\n ]*)",
             bygroups(Token.DIP.Text,  Token.DIP.Unit), '#pop:4'),
            (r"(\])", Token.DIP.Text, '#pop'),
            # units
            (r"([ ]*)([^#\n ]+)",
             bygroups(Token.DIP.Text, Token.DIP.String)),
            # end of definition
            (r'[ ]', Token.DIP.Text, '#pop:4'),
        ],
        'str_single' : [
            (r"[^'\\]+",               Token.DIP.String),
            (r"\\'",                   Token.DIP.String),
            (r"\\[^']",                Token.DIP.String),
            ("(')(,)",
             bygroups(Token.DIP.String, Token.DIP.Text), "#pop"),
            ("(')(\])",
             bygroups(Token.DIP.String, Token.DIP.Text), "#pop:2"),  # pop if in array
            ("(')", Token.DIP.String, "unit"),
        ],
        'str_double' : [
            (r'[^"\\]+',               Token.DIP.String),
            (r'\\"',                   Token.DIP.String),
            (r'\\[^"]',                Token.DIP.String),
            ('(")(,)',
             bygroups(Token.DIP.String, Token.DIP.Text), "#pop"),
            ('(")(\])',
             bygroups(Token.DIP.String, Token.DIP.Text), "#pop:2"),  # pop if in array
            ('"', Token.DIP.String, "unit"),
        ],
        'str_triple' : [
            (r'[^"\\]+',               Token.DIP.String),
            (r'\\"',                   Token.DIP.String),
            (r'\\[^"]',                Token.DIP.String),
            (r'"[^"]',                 Token.DIP.String),
            (r'""[^"]',                Token.DIP.String),
            (r'(""")(,)',
             bygroups(Token.DIP.String, Token.DIP.Text), "#pop"),
            (r'(""")(\])',
             bygroups(Token.DIP.String, Token.DIP.Text), "#pop:2"),  # pop if in array
            (r'"""', Token.DIP.String, "unit"),
        ],
        'expr_single' : [
            (r"\\({)",
             bygroups(Token.DIP.Expression)),
            (r'{',                     Token.DIP.Reference, 'reference_expr'),
            (r"[^'\\{]+",              Token.DIP.Expression),
            (r"\\'",                   Token.DIP.Expression),
            (r"\\[^']",                Token.DIP.Expression),
            ("(')(\))",
             bygroups(Token.DIP.Expression, Token.DIP.Text), "unit"),
        ],
        'expr_double' : [
            (r"\\({)",
             bygroups(Token.DIP.Expression)),
            (r'{',                     Token.DIP.Reference, 'reference_expr'),
            (r'[^"\\{]+',              Token.DIP.Expression),
            (r'\\"',                   Token.DIP.Expression),
            (r'\\[^"]',                Token.DIP.Expression),
            ('(")(\))',
             bygroups(Token.DIP.Expression, Token.DIP.Text), "unit"),
        ],
        'expr_triple' : [
            (r"\\({)",
             bygroups(Token.DIP.Expression)),
            (r'{',                     Token.DIP.Reference, 'reference_expr'),
            (r'[^"\\{]+',              Token.DIP.Expression),
            (r'\\"',                   Token.DIP.Expression),
            (r'\\[^"]',                Token.DIP.Expression),
            (r'"[^"]',                 Token.DIP.Expression),
            (r'""[^"]',                Token.DIP.Expression),
            (r'(""")(\))',
             bygroups(Token.DIP.Expression, Token.DIP.Text), "unit"),
        ],
        'reference' : [
            (r"{",                     Token.DIP.Reference, "#push"),
            (r"[^{}]+",                Token.DIP.Reference),
            (r"}\[",                   Token.DIP.Reference, "reference_slice"),
            (r"}",                     Token.DIP.Reference, "unit"),
        ],
        'reference_expr' : [
            (r"{",                     Token.DIP.Reference, "#push"),
            (r"[^{}]+",                Token.DIP.Reference),
            (r"}\[",                   Token.DIP.Reference, "reference_slice"),
            (r"(}:)([0-9sdef.]+)",
             bygroups(Token.DIP.Reference, Token.DIP.Slice), "#pop"),
            (r"}",                     Token.DIP.Reference, "#pop"),
        ],
        'reference_slice': [
            (r'[0-9.:]+',              Token.DIP.Slice),
            (r',',                     Token.DIP.Reference),
            (r'\]}',                   Token.DIP.Reference, '#pop:3'),
            (r'(\]:)([0-9sdef.]+)',
             bygroups(Token.DIP.Reference, Token.DIP.Slice), '#pop:2'),
            (r'\]',                    Token.DIP.Reference, 'unit'),
        ],
        'unit': [
            (r'([ ]+[^=#\n ]+|)',      Token.DIP.Unit, "#pop:6")
        ],
    }



import numpy as np
import pytest
from math import isclose
import os
import sys
sys.path.insert(0, 'src')

from scinumtools.units import *
    
def test_initialization():
    assert str(Magnitude(1))           == "1.000e+00"           # no error
    assert str(Magnitude(32, 0.3))     == "3.200(30)e+01"       # absolute error
    assert str(Magnitude(32, 3e-6))    == "3.20000000(30)e+01"    
    assert str(Magnitude(32, rele=10)) == "3.20(32)e+01"        # relative error
    
def test_values():
    
    m = Magnitude(30, 0.3)
    assert m.abse()  == 0.3
    assert m.rele()  == 1.0
    assert str(m.abse(0.5)) == "3.000(50)e+01"
    assert str(m.rele(30))  == "3.00(90)e+01"

def test_arithmetics():
    a = Magnitude(4, 0.01)
    b = Magnitude(1, 0.005)
    # addition
    assert str(a+b)    == "5.000(15)e+00"
    assert str(b+a)    == "5.000(15)e+00"
    assert str(a+3)    == "7.000(10)e+00"
    assert str(3+a)    == "7.000(10)e+00"
    # substraction
    assert str(a-b)    == "3.000(15)e+00"
    assert str(b-a)    == "-3.000(15)e+00"
    assert str(a-3)    == "1.000(10)e+00"
    assert str(3-a)    == "-1.000(10)e+00"
    assert str(-a)     == "-4.000(10)e+00"
    
    a = Magnitude(4, 0.05)
    b = Magnitude(7, 0.1)
    # multiplication
    assert str(a*b)    == "2.800(76)e+01"
    assert str(b*a)    == "2.800(76)e+01"
    assert str(a*5)    == "2.000(25)e+01"
    assert str(5*a)    == "2.000(25)e+01"
    
    a = Magnitude(12, 0.2)
    b = Magnitude(4, 0.1)
    # division
    assert str(a/b)    == "3.00(13)e+00"
    assert str(b/a)    == "3.33(14)e-01"
    assert str(a/6)    == "2.000(33)e+00"
    assert str(6/a)    == "5.000(85)e-01"
    # power
    assert str(a**2)   == "1.4400(40)e+02"
    
def test_numpy():
    
    a = Magnitude([12, 3], 0.2)
    b = Magnitude([2, 4], 0.1)
    assert str(a.value)  == "[12.  3.]"
    assert str(a.error)  == "[0.2 0.2]"
    assert str(a)        == "[1.200(20)e+01 3.00(20)e+00]"
    assert str(b)        == "[2.00(10)e+00 4.00(10)e+00]"
    assert str(a+b)      == "[1.400(30)e+01 7.00(30)e+00]"
    assert str(a*b)      == "[2.40(16)e+01 1.20(16)e+01]"
    assert str(a**2)     == "[1.4400(40)e+02 9.00(40)e+00]"

def test_quantities():
    
    m1 = Magnitude(12, 0.2)

    q1 = Quantity(m1, 'cm')
    q2 = Quantity(4, 'cm', abse=0.1)
    q3 = Quantity(4, 'cm', rele=10)
    q4 = Quantity(4, 'cm')
    q5 = Quantity([2, 4], 'cm', abse=0.1)
    
    assert str(q1)               == "Quantity(1.200(20)e+01 cm)"
    assert str(q2)               == "Quantity(4.00(10)e+00 cm)"
    assert str(q3)               == "Quantity(4.00(40)e+00 cm)"
    assert str(q4.abse(0.1))     == "Quantity(4.00(10)e+00 cm)"  # setting new Magnitude
    assert str(q4.rele(10))      == "Quantity(4.00(40)e+00 cm)"
    assert str(q4.abse(0.2))     == "Quantity(4.00(20)e+00 cm)"  # changing existing Mangitude error
    assert str(q4.rele(20))      == "Quantity(4.00(80)e+00 cm)"
    assert str(q5)               == "Quantity([2.00(10)e+00 4.00(10)e+00] cm)"
    
    assert str(q1+q2)  == "Quantity(1.600(30)e+01 cm)"
    assert str(q1*2)   == "Quantity(2.400(40)e+01 cm)"
    assert str(q1*q2)  == "Quantity(4.80(20)e+01 cm2)"
    assert str(q1**2)  == "Quantity(1.4400(40)e+02 cm2)"
    
    q = Quantity(30, 'cm', abse=0.3)
    assert q.abse()  == 0.3
    assert q.rele()  == 1.0
    assert str(q.abse(0.5)) == "Quantity(3.000(50)e+01 cm)"
    assert str(q.rele(30))  == "Quantity(3.00(90)e+01 cm)"


from reportlab.lib import colors
from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle
from reportlab.rl_config import defaultPageSize
from reportlab.lib.units import cm
from enum import Enum
import unicodedata
import numpy as np

from ...settings import ROOT_SOURCE

SAMPLE_STYLE_SHEET = getSampleStyleSheet()

FONT_NAME = 'Helvetica'
FONT_SIZE = 10
PAGE_HEIGHT=defaultPageSize[1]; 
PAGE_WIDTH=defaultPageSize[0]

TABLE_HEADER_STYLE = ParagraphStyle(
    "CustomStyle",
    parent=SAMPLE_STYLE_SHEET["Normal"],
    fontName=FONT_NAME,
    fontSize=FONT_SIZE,
    textColor=colors.saddlebrown
)

TABLE_BODY_STYLE = ParagraphStyle(
    "CustomStyle",
    parent=SAMPLE_STYLE_SHEET["Normal"],
    fontName=FONT_NAME,
    fontSize=FONT_SIZE,
)

NTYPES = [
    ('Declaration',                '#698D3F',),
    ('Definition',                 '#AABA78',),
    ('Declaration / Modification', '#D77F33',),
    ('Definition / Modification',  '#DDA15E',),
    ('Modification',               '#FBE974',),
    ('Injection',                  '#AB88BF',),
    ('Import',                     '#D0BCDC',),
]

PALETTE = {
    'c4': colors.saddlebrown, #'#',
    'node_name': '#E7CCB1', #colors.navajowhite, 
    'prop_name': '#FAEDCD', #colors.antiquewhite, #'#',
    'prop_value': '#FEFAE0', #colors.floralwhite, #'#',
}

TITLE = ParagraphStyle(
    name = 'Title',
    fontSize = 20,
    leading = 16,
    fontName='Times-Bold',
    spaceAfter=cm,
)

H0 = ParagraphStyle(
    name = 'Heading0',
    fontSize = 18,
    leading = 16,
    fontName='Times-Bold',
    spaceAfter=cm,
)

H1 = ParagraphStyle(
    name = 'Heading1',
    fontSize = 18,
    leading = 16,
    fontName='Times-Bold',
    spaceAfter=cm,
)

H2 = ParagraphStyle(
    name = 'Heading2',
    fontSize = 16,
    leading = 14,
    fontName='Times-Bold',
    spaceAfter=cm,
)

def Title(name:str=''):
    return f"<a name=\"SECTION_{name}\"></a>{name}"

def Target(key:str, name:str=''):
    return f"<a name=\"{key}\"></a>{name}"
    
def Link(key:str, name:str=''):
    return f"<a href=\"#{key}\" color=\"blue\">{name}</a>"
    
def HighlightReference(text:str):
    text = text.replace("{","<font color='orange'>{")
    text = text.replace("}","}</font>")
    return text
    
def ColumnWidths(ratios: list):
    return list(np.array(ratios)*(PAGE_WIDTH-5.2*cm))


from reportlab.platypus import Paragraph, Table, Spacer, PageBreak
from reportlab.lib.units import inch
import numpy as np
import re

from .settings import *
from ..settings import DocsType
from ...nodes import Node, BooleanNode, IntegerNode, FloatNode, StringNode, ModNode, ImportNode
from ...settings import Order, Sign, Keyword, EnvType

class NodeSection:
    
    data: list
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
        
    def __init__(self, data):
        self.data = data
            
    def parse_node(self, name, node):
    
        # define color
        color = NTYPES[node.ntype][1]
    
        # define table style
        TABLE_STYLE = [
            # whole grid
            #('GRID',       (0,0), (-1,-1),  0.5,     colors.goldenrod), 
            ('GRID',       (1,1), (-1,-1),  0.5, PALETTE['prop_name']),
            ('FONTNAME',   (0,0), (-1,-1),  TABLE_BODY_STYLE.fontName),
            ('FONTSIZE',   (0,0), (-1,-1),  TABLE_BODY_STYLE.fontSize),
            # top panel
            ('FONTSIZE',   (0,0), (-1,-1),  TABLE_HEADER_STYLE.fontSize),
            ('TEXTCOLOR',  (0,0), (-1,0),   PALETTE['c4']),   
            ('BACKGROUND', (0,0), (-1,0),   PALETTE['node_name']),
            ('ALIGN',      (-2,0),(-1,0),   'RIGHT'),
            ('BACKGROUND', (1,1), (1,-1),   PALETTE['prop_name']),  
            ('BACKGROUND', (2,1), (-1,-1),  PALETTE['prop_value']),   
            ('SPAN',       (1,0), (2,0)     ),                            # parameter name
            ('BACKGROUND', (0,0), (0,-1),   color),                       # instance type
        ]
        
        # construct a node table
        node_target = Target(node.target)
        source_link = Link(node.link_source, f"{node.source[0]}:{node.source[1]}")
        injection_link = Link(node.link_injection, " | injected") if node.injection else ''
        import_link = Link(node.link_import, " | imported") if node.imported else ''
        dtype = node.dtype
        if node.constant:
            dtype = f"constant {dtype}"
        data = [
            ['', Paragraph(f"{node_target}{source_link}{injection_link}{import_link}"), '', dtype],
        ]
        if node.value:
            TABLE_STYLE.append(('SPAN', (2,len(data)), (-1,len(data)) ))
            data.append(['', 'Value:',        Paragraph(node.value, TABLE_BODY_STYLE)])
        if node.unit:
            TABLE_STYLE.append(('SPAN', (2,len(data)), (-1,len(data)) ))
            data.append(['','Unit:',          Paragraph(node.unit, TABLE_BODY_STYLE)])
        if node.condition:
            TABLE_STYLE.append(('SPAN', (2,len(data)), (-1,len(data)) ))
            data.append(['','Condition:',     Paragraph(HighlightReference(node.condition), style=TABLE_BODY_STYLE)])
        if node.tags:
            TABLE_STYLE.append(('SPAN', (2,len(data)), (-1,len(data)) ))
            data.append(['','Tags:',          Paragraph(", ".join(node.tags), TABLE_BODY_STYLE)])
        if node.options: 
            TABLE_STYLE.append(('SPAN', (2,len(data)), (-1,len(data)) ))
            data.append(['','Options:',       Paragraph(", ".join(node.options), TABLE_BODY_STYLE)])
        if node.dformat:
            TABLE_STYLE.append(('SPAN', (2,len(data)), (-1,len(data)) ))
            data.append(['','Format:',        Paragraph(node.dformat, TABLE_BODY_STYLE)])
        if node.description:
            TABLE_STYLE.append(('SPAN', (2,len(data)), (-1,len(data)) ))
            data.append(['','Description:',   Paragraph(node.description, TABLE_BODY_STYLE)])
            
        return Table(
            data, style=TABLE_STYLE, hAlign='LEFT', 
            colWidths=ColumnWidths([0.01,0.2, 0.50, 0.29])
        )

    def parse(self):
        blocks = []
        blocks.append(Paragraph(Title(f"Parameter nodes"), H2) )
        names = list(self.data.keys())
        names.sort()
        for name in names:
            pdata = self.data[name]
            blocks.append(Spacer(1,0.1*inch))
            blocks.append(Paragraph(f"<strong>"+Target(pdata.target, name)+"</strong>"))
            blocks.append(Spacer(1,0.1*inch))
            for node in pdata.nodes:
                blocks.append(self.parse_node(name, node))
        blocks.append(Spacer(1,0.2*inch))
        return blocks


from pygments.style import Style
from pygments.token import Token

class StyleLexer(Style):

    styles = {
        # node components
        Token.DIP.Indent:                 '#BDBDBD',        # indent tag
        Token.DIP.Keyword:                '#757575',        # directives $ and !
        Token.DIP.Name:                   'bold #616161',   # node name
        Token.DIP.Type:                   '#616161',        # node type
        Token.DIP.Dimension:              '#9E9E9E',        # node dimension
        Token.DIP.String:                 '#558B2F',        # value string
        Token.DIP.Number:                 '#B71C1C',        # value number
        Token.DIP.Boolean:                'bold #0D47A1',   # value bool
        Token.DIP.Reference:              '#E64A19',        # reference
        Token.DIP.Slice:                  '#FF8A65',        # reference slice     
        Token.DIP.Expression:             '#8D6E63',        # expression
        Token.DIP.Unit:                   '#607D8B',        # unit
        Token.DIP.Comment:                'italic #FB8C00', # comment

        # schema components
        Token.DIP.Tag:                    '#0097A7',        # tag

        # other components
        Token.Token.Error:            'border:#F44336', # missing highlight

        # default highlighting
        Token.Whitespace:                "#bbbbbb",
        Token.Comment:                   "italic #3D7B7B",
        Token.Comment.Preproc:           "noitalic #9C6500",

        #Keyword:                   "bold #AA22FF",
        Token.Keyword:                   "bold #008000",
        Token.Keyword.Pseudo:            "nobold",
        Token.Keyword.Type:              "nobold #B00040",

        Token.Operator:                  "#666666",
        Token.Operator.Word:             "bold #AA22FF",

        Token.Name.Builtin:              "#008000",
        Token.Name.Function:             "#0000FF",
        Token.Name.Class:                "bold #0000FF",
        Token.Name.Namespace:            "bold #0000FF",
        Token.Name.Exception:            "bold #CB3F38",
        Token.Name.Variable:             "#19177C",
        Token.Name.Constant:             "#880000",
        Token.Name.Label:                "#767600",
        Token.Name.Entity:               "bold #717171",
        Token.Name.Attribute:            "#687822",
        Token.Name.Tag:                  "bold #008000",
        Token.Name.Decorator:            "#AA22FF",

        Token.String:                    "#BA2121",
        Token.String.Doc:                "italic",
        Token.String.Interpol:           "bold #A45A77",
        Token.String.Escape:             "bold #AA5D1F",
        Token.String.Regex:              "#A45A77",
        #Token.String.Symbol:             "#B8860B",
        Token.String.Symbol:             "#19177C",
        Token.String.Other:              "#008000",
        Token.Number:                    "#666666",

        Token.Generic.Heading:           "bold #000080",
        Token.Generic.Subheading:        "bold #800080",
        Token.Generic.Deleted:           "#A00000",
        Token.Generic.Inserted:          "#008400",
        Token.Generic.Error:             "#E40000",
        Token.Generic.Emph:              "italic",
        Token.Generic.Strong:            "bold",
        Token.Generic.EmphStrong:        "bold italic",
        Token.Generic.Prompt:            "bold #000080",
        Token.Generic.Output:            "#717171",
        Token.Generic.Traceback:         "#04D",

        Token.Error:                     "border:#FF0000"
    }

def pygments_monkeypatch_style(mod_name, cls):
    import sys
    import pygments.styles
    cls_name = cls.__name__
    mod = type(__import__("os"))(mod_name)
    setattr(mod, cls_name, cls)
    setattr(pygments.styles, mod_name, mod)
    sys.modules["pygments.styles." + mod_name] = mod
    from pygments.styles import STYLE_MAP
    STYLE_MAP[mod_name] = mod_name + "::" + cls_name



import numpy as np
import pytest
from math import isclose
import os
import sys
sys.path.insert(0, 'src')

from scinumtools.units import Quantity, Unit 
from scinumtools.materials import *

def test_preprocessing():
    
    material = Material()
    with MaterialSolver(material.atom) as ms:
        
        # implicit multiplication and addition
        assert ms.preprocess('1.0 <H2O>')             == "1.0 * <H2O>"
        assert ms.preprocess('0.5 <H2O> 0.5 <NaCl>')  == "0.5 * <H2O> + 0.5 * <NaCl>"
        assert ms.preprocess('<H2O>')                 == "<H2O>"
        assert ms.preprocess('<H2O> <NaCl>')          == "<H2O> + <NaCl>"
        
def test_solver():
    
    material = Material()
    with MaterialSolver(material.atom) as ms:
        
        # single atom parsing
        assert str(ms.solve('<H2O>'))                 == "Material(1.0 H2O)"
        assert str(ms.solve('0.5 <H2O>'))             == "Material(0.5 H2O)"
        assert str(ms.solve('0.2 <H2O> 0.8 <NaCl>'))  == "Material(0.2 H2O; 0.8 NaCl)"
        assert str(ms.solve('0.7808 <N2> 0.2095 <O2> 0.0093 <Ar> 0.0004 <CO2>')) == \
                    "Material(0.7808 N2; 0.2095 O2; 0.0093 Ar; 0.0004 CO2)"



from reportlab.platypus import Paragraph, Table, Spacer
from reportlab.lib.units import inch
import numpy as np

from .settings import *
from ..settings import DocsType

class ParametersSection:
    
    names: list
    data: list
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
        
    def __init__(self, data):
        self.data = data
        self.names = list(self.data.keys())
        self.names.sort()
    
    def parse_table(self):
        TABLE_STYLE = [
            ('GRID',       (0,1), (-1,-1),  0.5, PALETTE['prop_name']),
            ('BACKGROUND', (0,0), (0,0),   PALETTE['node_name']),  
            ('BACKGROUND', (0,1), (-1,-1),   PALETTE['prop_value']),  
            ('BACKGROUND', (1,0), (1,0),   NTYPES[0][1]),  
            ('BACKGROUND', (2,0), (2,0),   NTYPES[1][1]),  
            ('BACKGROUND', (3,0), (3,0),   NTYPES[2][1]),  
            ('BACKGROUND', (4,0), (4,0),   NTYPES[3][1]),  
            ('BACKGROUND', (5,0), (5,0),   NTYPES[4][1]),  
            ('BACKGROUND', (6,0), (6,0),   NTYPES[5][1]),  
            ('BACKGROUND', (7,0), (7,0),   NTYPES[6][1]),  
        ]
    
        data = [['Property name', '#', '#', '#', '#', '#', '#', '#']]
        for name in self.names:
            item = self.data[name]
            #p = Paragraph(AnchorLink(AnchorType.PARAM, item.name))
            p = Paragraph(Link(item.target, item.name))
            counts = ['' if c==0 else c for c in item.counts]
            data.append([p]+counts)
        return Table(data, style=TABLE_STYLE, hAlign='LEFT', colWidths=ColumnWidths([0.72, 0.04, 0.04, 0.04, 0.04, 0.04, 0.04, 0.04]))
        
    def parse(self):
        blocks = []       
        blocks.append(Paragraph(Title(f"Parameter list"), H2))
        blocks.append(self.parse_table())
        blocks.append(Spacer(1,0.2*inch))
        return blocks


import numpy as np
import os
import pytest
from PIL import Image
from io import StringIO 
import sys
sys.path.insert(0, 'src')

from scinumtools import ImageMetadata, Metadata

class Capturing(list):
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self
    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio    # free up some memory
        sys.stdout = self._stdout
        
@pytest.fixture()
def temp_file():
    dir_temp = "tmp"
    name_temp = "thumbnail.png"
    file_name = f"{dir_temp}/{name_temp}"
    if not os.path.isdir(dir_temp):
        os.mkdir(dir_temp)
    data = np.full((10,10),1)
    with Image.fromarray(data.astype('uint8')) as im:
        im.save(file_name)
    return file_name
    
def test_read_exif(temp_file):
    assert os.path.isfile(temp_file)
    
    with ImageMetadata(temp_file) as im:
        im.set(Metadata.DATETIME,'2024-04-10 18:00:00')

    with ImageMetadata(temp_file) as im:
        assert im.get(Metadata.DATETIME) == '2024-04-10 18:00:00'

def test_print(temp_file):
    assert os.path.isfile(temp_file)
    
    with Capturing() as output:
        with ImageMetadata(temp_file) as im:
            im.set(Metadata.DATETIME,'2024-04-10 18:00:00')
            im.print()
    assert output == [
        'DateTime: 2024-04-10 18:00:00'
    ]


from .node_base import BaseNode

class ConditionNode(BaseNode):
    keyword: str = 'condition'

    @staticmethod
    def is_node(parser):
        parser.kwd_condition()
        if parser.is_parsed('kwd_condition'):
            parser.part_value()
            parser.part_comment()
            return ConditionNode(parser)
            
    def parse(self, env):
        env.nodes[-1].condition = self.value_expr
        return None



import csv
import json

from .node_base import BaseNode
from .parser import Parser
from . import BooleanNode, IntegerNode, FloatNode, StringNode
from ..settings import Sign

class TableNode(BaseNode):
    keyword: str = 'table'
    
    @staticmethod
    def is_node(parser):
        if parser.keyword=='table':
             parser.part_dimension()
             parser.part_equal()
             if parser.is_parsed('part_equal'): # definition
                 parser.part_value()  
             else:
                 parser.defined = True  # declaration
             parser.part_units()    
             parser.part_comment()
             return TableNode(parser)
         
    def parse(self, env):
        lines = self.value_raw.split(Sign.NEWLINE)
        # Parse nodes from table header
        table = []
        dimension = []
        while len(lines)>0:
            line = lines.pop(0)
            if line.strip()=='':
                break
            # Parse node parameters
            parser = Parser(
                code=line,
                source=self.source
            )
            parser.part_name()      # parse node name
            parser.part_type()      # parse node type
            parser.part_dimension() # parse node type dimension
            parser.part_units()     # parse node units
            if not parser.is_empty():
                raise Exception(f"Incorrect header format: {line}")
            # Initialize actual node
            types = {
                'bool':  BooleanNode,
                'int':   IntegerNode,
                'float': FloatNode,
                'str':   StringNode,
            }
            # Create column
            if parser.keyword in types:
                node = types[parser.keyword](parser)
                node.value_raw = []
                table.append(node)
            else:
                raise Exception(f"Incorrect format or missing empty line after header: {self.code}")
        # Remove whitespaces from all table rows
        for l,line in enumerate(lines):
            lines[l] = line.strip()
            if line=='': del lines[l]
        # Read table and assign its values to the nodes
        ncols = len(table)
        csvtab = csv.reader(lines, delimiter=' ')
        for row in csvtab:
            if len(row)>ncols or len(row)<ncols:
                raise Exception(f"Number of header nodes does not match number of table columns: {ncols} != {len(row)}")
            for c in range(ncols):
                if table[c].dimension:
                    table[c].value_raw.append(json.loads(row[c]))
                else:
                    table[c].value_raw.append(row[c])
        # set additional node parameters
        nodes_new = []
        for node in table:
            nvalues = len(node.value_raw)
            node.dimension = [(nvalues,nvalues)]
            node.name = self.name + Sign.SEPARATOR + node.name
            node.indent = self.indent
            nodes_new.append(node)
        return nodes_new



from typing import List, Dict, Union
from dataclasses import dataclass, field
import copy

from ..settings import Sign
from .list_nodes import NodeList

@dataclass
class EnvSource:
    name: str            # this source name
    path: str            # source filename
    code: str            # source code
    parent: tuple = None   # parent source
    nodes: NodeList = None      # list of nodes (only for remote sources)
    sources: 'SourceList' = None  # list of sources (only for remote sources)
    
@dataclass
class SourceList:
    sources: Dict   = field(default_factory = dict)  # custom reference sources
    
    def __len__(self):
        return len(self.sources)
    
    def __getitem__(self, name: str):
        if name not in self.sources:
            raise Exception("Source with the following name does not exist:", name)
        return self.sources[name]

    def __setitem__(self, name: str, source: EnvSource):
        self.sources[name] = source

    def __contains__(self, item):
        return item in self.sources
    
    def items(self):
        return self.sources.items()
    
    def keys(self):
        return self.sources.keys()
    
    def copy(self):
        """ Copy a new object from self
        """
        return copy.deepcopy(self)
        
    def append(self, name:str, path:str, code:str, parent:tuple = None):
        """ Append a new source

        :param str name: Name of a new source
        :param source: Source can be either a DIP object or a text
        :param str path: Source path
        """
        if name in self.sources:
            raise Exception("Reference source alread exists:", name)
        self.sources[name] = EnvSource(name, str(path), code,  parent)

    def query(self, query:str):
        """ Select sources according to a query

        :param str query: Source selection query
        """
        if query==Sign.WILDCARD:   # return all sources
            return self.sources
        else:                     # return particular source
            if query not in self.sources:
                raise Exception("Requested source does not exists:", query)
            return {query: self.sources[query]}


from bs4 import BeautifulSoup
import json
from pathlib import Path
from pygments import highlight
from pygments.lexers import PythonLexer
from pygments.formatters import get_formatter_by_name
from pygments.styles import get_style_by_name
from pygments.token import STANDARD_TYPES, Token
from pygments.formatters import HtmlFormatter
import re

from .settings import *
from .rst_parser import ParseRST
from ..item_parameter import ParType
from ..documentation import Documentation
from ...settings import Sign, ROOT_SOURCE

class SettingsSection:
        
    docs: Documentation
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
    
    def __init__(self, docs: Documentation, **kwargs):
        self.docs = docs
        self.rst = ParseRST()
        
    def build_units(self):
        
        self.rst.title(1,'Units')
        
        data = [['Name','Value','Units','Source']]
        for unit in self.docs.units:
            data.append([
                unit.name, 
                f"``{unit.value}``", 
                f"``{unit.units}``", 
                self.rst.link(unit.link_source, f"{unit.source[0]}:{unit.source[1]}", source=True)
            ])
        self.rst.table(data, header=True)
        
    def build_sources(self):
        
        self.rst.title(1,'Sources')
        
        for sdata in self.docs.sources:
            
            if sdata.parent:
                data = [[sdata.name,'']]
            else:
                data = [[sdata.name,'']] 
            merge = [(0,0,0,1)]
        
            file_name = Path(sdata.path).name
            data.append(['File:',file_name])
        
            if sdata.parent:
                data.append(['Source:', self.rst.link(sdata.link_source, f"{sdata.parent[0]}:{sdata.parent[1]}", source=True)])
        
            self.rst.target(sdata.target)
            self.rst.table(data, merge=merge, header=True)
            
            if sdata.code:
                self.rst.code(sdata.code,'DIP',linenos=True,caption=file_name) #,name=sdata.target)
        
    def build(self):
        
        self.build_units()
        
        self.build_sources()
        
        return self.rst.parse()